{-# OPTIONS -cpp #-}
-- #ignore-exports

{- | Routines to transform 'Parser.SimpleRule's and productions into 
     Chomsky Normal Form.

'Parser.SimpleRule's such as

> f(g(g(X)),X) => c(X) .

are transformed into equivalent 'Parser.FullRule's having 'Term' patters of 
maximum depth 1 and 'Variable's occurring at most once.  For instance, the 
Chomsky Normal Form of the above 'Parser.SimpleRule' is

> | L : f(!2, X!1), !2 : g(!3), !3 : g(X!0), @!Inters : cons(X!0, X!1) => L : c(X!0) .

Similarly, productions such as

> @a : f(g(@b)) .

are transformed into

> @a : f(%0) .
> %0 : g(@b) .

-}

module ChomskyNormalForm
    ( normalizeSimpleRule,
      normalizeSimpleRules,
      normalizeProductions
    )
where

import Control.Monad.State
import qualified Data.Map as Map

import Syntax

-- * Basic Operations

-- | State for the normalization of 'Parser.SimpleRule's.
data RSt = 
     RSt 
     { -- | Used to generate fresh variables.
       counterR :: ! Int
       -- | Stores all the aliases for a variable.
     , aliases :: ! (Map.Map Variable [Variable])
     }

type CNFRSt = State RSt

-- | State for the normalization of productions.
data GSt = 
    GSt 
    { -- | Used to generate fresh 'Label's. 
      counterG :: ! Int 
      -- | Productions in CNF
    , cnfProductions :: ! [(Label, [Term])]
    }
type CNFGSt = State GSt

-- | Generate a fresh 'Int' and increment 'counterR'.
newId :: CNFRSt Int
newId =
    do
    i <- gets counterR
    modify (\st -> st{counterR = succ i})
    return i

-- | Generate a fresh for-internal-use-only 'Variable'.
newVariable :: CNFRSt Variable
newVariable =
    do
    i <- newId
    return $ Variable $ '!' : show i

-- | Generate fresh 'Label'.
newLabel :: CNFGSt Label
newLabel =
    do
    i <- gets counterG
    modify (\st -> st{ counterG = succ i})
    return $ Label $ "%" ++ show i

-- | Add extra productions to 'GSt'.
writeProduction :: (Label,[Term]) -> CNFGSt ()
writeProduction prod =
    modify (\st -> st{ cnfProductions = prod : cnfProductions st })

-- | Generte a fresh alias for a 'Variable', and store the alias in the 
--   'aliases' map.
newAlias :: Variable -> CNFRSt Variable
newAlias v@(Variable vname) =
    do
    i <- newId
    st <- get
    let v' = Variable $ vname ++ "!" ++ show i
    put st{ aliases = Map.insertWith (++) v [v'] (aliases st) }
    return v'

-- * Clause Normalization

-- | Generate a list of intersection 'Clause's for each distinct pair of
--   aliases for a 'Variable'.
aliasesIntersections :: CNFRSt [Clause]
aliasesIntersections =
    do
    toIntersect <- gets $ Map.elems . aliases
    let mkIntersect :: [Variable] -> [Clause]
        mkIntersect [] = []
	mkIntersect (v:vs) = 
	    [ Clause (LLab intersectionLabel) 
                  (App intersectionHead [Var v, Var v'] )
	      | v' <- vs ] ++ 
	    mkIntersect vs
    return $ concatMap mkIntersect toIntersect

-- | Transform a 'Clause' having an arbitrary 'Term' as pattern into
--   a list of 'Clause's where pattern have maximum depth 1.
flattenClause :: Clause -> CNFRSt [Clause]
flattenClause cl@(Clause _ MinTerm) = return [cl]
flattenClause cl@(Clause _ (Lab _)) = return [cl]
flattenClause cl@(Clause _ (Var _)) = return [cl]
flattenClause (Clause loc (App h ts)) =
    do
    let aux :: Term -> CNFRSt (Term,[Clause])
	aux t@MinTerm = return (t,[])
	aux t@(Lab _) = return (t,[])
	aux t@(Var _) = return (t,[])
	aux t@(App _ _) =
	    do
	    v <- newVariable
	    cls <- flattenClause (Clause (LVar v) t)
	    return (Var v,cls)
    res <- mapM aux ts
    let (ts',cls') = unzip res
    return $ (Clause loc (App h ts')) : concat cls'

-- | Transform a 'Clause' having an arbitrary 'Term' as pattern into
--   a 'Clause' list having only patterns with maximum depth 1, and each
--   'Variable' occurring at most once in each 'Clause'.
normalizeClause :: Clause -> CNFRSt [Clause]
normalizeClause (Clause loc term) =
    do
    let -- Replace every 'Variable' with an alias.
        renameTerm :: Term -> CNFRSt Term
	renameTerm MinTerm = return MinTerm
        renameTerm (Var v) =
	    do
	    v' <- newAlias v
	    return $ Var v'
	renameTerm t@(Lab _) = return t
	renameTerm (App h ts) =
	    do
	    ts' <- mapM renameTerm ts
	    return $ App h ts'
    -- Remove multiple 'Variable' occurrences.
    term'  <- renameTerm term
    -- Bring everything to depth <= 1.
    cls    <- flattenClause $ Clause loc term'
    -- Add the intersection requirements between each alias.
    intCls <- aliasesIntersections
    return $ cls ++ intCls

-- | The 'Subst' associating a 'Variable' with its first alias.
aliasSubst :: CNFRSt Subst
aliasSubst =
    do
    al <- gets aliases
    return $ Subst [ (v,Var v') | (v,(v':_)) <- Map.toList al ]

-- * Main Routines

-- | Transform a 'Parser.SimpleRule' into a 'Parser.FullRule'.
normalizeSimpleRule :: Rule -> Rule
normalizeSimpleRule rule =
    let norm :: Rule -> CNFRSt Rule
	norm ([clause] :=> goal) =
	    do
            -- Put the first 'Clause' in CNF.
	    clause' <- normalizeClause clause
            -- Apply the 'Subst' to the goal,
	    subst <- aliasSubst
	    let goal' = applySubst subst goal
	    return $ clause' :=> goal'
        norm _ = error "normalizeSimpleRule: not a SimpleRule!"

        initialState :: RSt
        initialState = RSt { counterR = 0 , aliases = Map.empty }
	in evalState (norm rule) initialState

normalizeSimpleRules :: [Rule] -> [Rule]
normalizeSimpleRules = map normalizeSimpleRule

-- | Flatten a production.
flattenProduction :: (Label, [Term]) -> CNFGSt ()
flattenProduction (label, terms) =
    let -- Top level
        normTerm1 :: Term -> CNFGSt Term
	normTerm1 t@(App h ts) =
	    if isPlain t
	    then return t
	    else do ts' <- mapM normTerm ts ; return $ App h ts'
	normTerm1 t = return t 
	-- Deep level
	normTerm :: Term -> CNFGSt Term
	normTerm (App h ts) =
	    do
	    l <- newLabel
	    let lab = Lab l
	    ts' <- mapM normTerm ts
	    writeProduction (l, [App h ts'])
	    return lab
	normTerm t = return t
	in do
	   terms' <- mapM normTerm1 terms
	   writeProduction (label, terms')

-- | Normalize productions.
normalizeProductions :: [(Label, [Term])] -> [(Label, [Term])]
normalizeProductions ps =
    let normProds :: [(Label, [Term])] -> CNFGSt ()
	normProds = mapM_ flattenProduction
	initialState :: GSt
	initialState = GSt { counterG = 0 , cnfProductions = [] }
	in cnfProductions $ execState (normProds ps) initialState
