{-# OPTIONS -cpp #-}
-- #ignore-exports
module Grammar
( -- * Types
  Grammar(),
  -- ** Common Reader and State Monads
  GrReader,
  GrState,
  -- * Basic Constants and Operations
  emptyGr,
  addToGr,
  changed,
  clearGr,
  changeGr,
  -- * Queries
  -- ** Boolean Queries
  knowsLabel,
  plainElem, 
  pureElem,
  isEmptyPlainTerm,
  -- ** Search Queries
  -- *** Simple Queries
  plainTermsOfLabel,
  plainTermsOfLabel_h,
  labelsOfPlainTerm,
  -- *** Pattern Queries
  matchPattern_h,
  matchPattern_hlv,
  matchPattern_hvl,
  -- ** Global Queries
  knownLabels,
  allKnownInfo,
  -- * Pretty Printing
  pprGr,
  pprSaveGr,
  -- * Special Operations
  closeUnderTransitivity, 
  closeUnderIntersection,
  computeIntersection,
  samePlainTermsOfLabels,
  joinLabels,
  numberOfPlainTermsOfLabel
)
where

-- Imports
import Control.Monad.State
import Control.Monad.Reader
import Data.Maybe
import Data.List (sortBy, intersperse)
import qualified Data.Set as Set
import qualified Data.Map as Map

-- import Debug.Trace

import Renamer
import qualified Syntax (Label(Label))  -- only needed for pprSaveGr
import Utility

-- * Types

-- | Terms, ordered using direct lexicographic ordering.
--   That is, @f(a, a) < f(a, b) < f(b, a) < f(b, b)@
newtype TermDir = TermDir Term deriving (Eq, Show, Ord)

-- | Terms, ordered using inverse lexicographic ordering.
--   That is, @f(a, a) < f(b, a) < f(a, b) < f(b, b)@
newtype TermInv = TermInv Term deriving (Eq, Show)

-- Ensured by type checking: term head arity is constant.
instance Ord TermInv where
    compare (TermInv t1) (TermInv t2) =
	case (t1,t2) of
	    (App h1 ts1, App h2 ts2) ->
	        case compare h1 h2 of
	            EQ -> compare (reverse ts1) (reverse ts2)
	            res -> res
            _ -> compare t1 t2

-- | The data structure that holds all the information about the grammar.
data Grammar = Gr
    { -- | The label-to-set direct map.
      labelToTerms    :: ! (Map.Map Label (Set.Set Term))
      -- | Maps every plain 'Term' to the 'Label's of the sets the 'Term' 
      --   occurs in.
    , termDirToLabels :: ! (Map.Map TermDir (Set.Set Label))
      -- | Maps every plain 'Term' to the 'Label's of the sets the 'Term' 
      --   occurs in.
    , termInvToLabels :: ! (Map.Map TermInv (Set.Set Label))
      -- | The pairs of labels whose sets are still not known to intersect.
    , unknownIntersection :: ! (Set.Set (Label,Label))
      -- | Flag set whenever the grammar is changed.
    , changed :: ! Bool
      -- | User labels we joined.
    , joined :: [(Label, Label)]
    }

-- ** Reader and State Monads

-- | Read-only 'Grammar' action
type GrReader = Reader Grammar
-- | Read-write 'Grammar' action
type GrState = State Grammar

-- | Transform a read-only action in a read-write one (which actually
--   does not write anything).
actGr :: GrReader a -> GrState a
actGr action = gets $ runReader action

-- * Basic Constants and Operations

-- | An empty 'Grammar', except for 'intersectionLabel'.
emptyGr :: Grammar
emptyGr = let zeroGr = Gr{ labelToTerms        = Map.empty
			 , termDirToLabels     = Map.empty
			 , termInvToLabels     = Map.empty
			 , unknownIntersection = Set.empty
                         , changed             = False
                         , joined              = []
			 }
	      in execState (addToGr intersectionLabel
		           (App intersectionHead
		            [Lab intersectionLabel, Lab intersectionLabel]))
                 zeroGr

-- | Add a /plain/ 'Term' to the set of a 'Label', updating all the maps.
--   Return whether the 'Term was already in the 'Grammar'.
addToGr :: Label -> Term -> GrState Bool
addToGr lab term =
    do
    touchLabel lab
    alreadyThere <- actGr $ plainElem lab term
    unless alreadyThere $
         (do
          gr <- get
          -- The new maps.
          let fmL = augmentSet lab            term (labelToTerms gr) 
	  let fmD = augmentSet (TermDir term) lab  (termDirToLabels gr)
	  let fmI = augmentSet (TermInv term) lab  (termInvToLabels gr)
          put $ gr{ labelToTerms    = fmL
	          , termDirToLabels = fmD
		  , termInvToLabels = fmI
		  }
          changeGr
         )
    return alreadyThere

-- | Clear the 'changed' flag.
clearGr :: GrState()
clearGr = modify $ \gr -> gr{changed = False}

-- | Set the 'changed' flag.
changeGr :: GrState ()
changeGr = modify $ \gr -> gr{changed = True}

-- | If 'Label' is not present in the 'Grammar', updates 
--   'unknownIntersection' to schedule the check for new intersections.
--   Also assigns an empty set to the 'Label'.
touchLabel :: Label -> GrState ()
touchLabel lab =
    do
    know <- actGr $ knowsLabel lab
    unless know 
        (do
         allLabels <- actGr $ knownLabels
	 let newPairs = [ (lab,l2) | l2 <- lab : allLabels ] ++
	                [ (l2,lab) | l2 <- allLabels ]
	 modify $ \gr -> 
            gr{ labelToTerms = Map.insert lab Set.empty (labelToTerms gr)
	      , unknownIntersection = unknownIntersection gr
	        `Set.union`  Set.fromList newPairs
	      }
         changeGr
        )

-- * Queries
-- ** Boolean Queries

-- | Checks whethera 'Label' is present is the 'Grammar'.
knowsLabel :: Label -> GrReader Bool
knowsLabel lab =
    do
    ltt <- asks labelToTerms
    return $ isJust $ lab `Map.lookup` ltt

-- | Check whether a /plain/ 'Term' belongs to the set of the given 'Label'.
plainElem :: Label -> Term -> GrReader Bool
plainElem lab term =
    -- ASSUMPTION Closure under transitivity.
    do
    ltt <- asks labelToTerms
    return $ maybe False (Set.member term) $ lab `Map.lookup` ltt

-- | Check whether a /pure/ 'Term' belongs to the set of the given 'Label'. 
--   The 'Term' can be arbitrarily deep.
pureElem :: Label -> Term -> GrReader Bool
pureElem lab (App h ts) =
    do
    terms <- plainTermsOfLabel lab
    mor [ mand [ pureElem lab' t | 
                 (Lab lab', t) <- zip ts' ts
               ] | 
          App h' ts' <- terms,  -- Since terms are plain, ts' are labels.
          h' == h
          -- Ensured by type checking: length ts' == length ts .
        ]
pureElem _ _ = error "pureElem: term must be a pure term"

-- | Check whether a /plain/ term can be expanded into a /pure/ term, or
--   instead it denotes an \"empty\" term set.
isEmptyPlainTerm :: Term -> GrReader Bool
isEmptyPlainTerm (App _ ts) = mor [ isEmptyPlainTerm t | t <- ts ]
isEmptyPlainTerm t@(Lab _)  = liftM not $ 
    plainElem intersectionLabel $ App intersectionHead [t,t]
isEmptyPlainTerm _ = error "isEmptyPlainTerm"

-- ** Search Queries
-- *** Simple Queries

-- | Return all the occurrences of a /plain/ 'Term'.
labelsOfPlainTerm :: Term -> GrReader [Label]
labelsOfPlainTerm term =
    do
    -- assert
    if isPlain term then return () else error "labelsOfPlainTerm"
    tdtl <- asks termDirToLabels
    let listTsetL = Map.toAscList $ split_GE tdtl (TermDir term)
    let isGood = (== (TermDir term))
    let listTsetL' = takeWhile (isGood . fst) listTsetL
    let result = concatMap (Set.toList . snd) listTsetL'
    -- assert
    case term of (Lab _) -> if result == [] 
			    then error "labelsOfPlainTerm2"
			    else return ()
		 _       -> return ()
    return result

-- | Return all the set of /plain/ 'Term's of a 'Label'.
plainTermsOfLabel :: Label -> GrReader [Term]
plainTermsOfLabel l =
    do
    ltt <- asks labelToTerms
    return $ maybe [] Set.toList $ l `Map.lookup` ltt

-- | Return all the set of /plain/ 'Term's of a 'Label' matching @h(X,...,Y)@.
--   Arity can be @\/=@ 2.
plainTermsOfLabel_h :: Label -> Head -> GrReader [Term]
plainTermsOfLabel_h l h =
    do
    ltt <- asks labelToTerms
    let from, to :: Term
	from = App h []
	to   = App (succHead h) []
    return $ maybe [] (Set.toList . clampSet from to) $ l `Map.lookup` ltt

-- *** Pattern Queries

-- | Return all the matches for a @h(X,...,Y)@ pattern. Arity can be @\/=@ 2.
matchPattern_h :: Head -> GrReader [(Label,Term)]
matchPattern_h h =
    do
    tDtl <- asks termDirToLabels
    let listTsetL = Map.toAscList $ split_GE tDtl (TermDir (App h []))
	isGood (TermDir (App h' _)) = h' == h
	isGood _ = False
	listTsetL' = takeWhile (isGood . fst) listTsetL
	expand (k,e) = [ (l,t) | let TermDir t = k , l <- Set.toList e ] 
    return $ concatMap expand listTsetL'

-- | Return all the matches for a @h(\@lab,Y)@ pattern. Only arity 2.
matchPattern_hlv :: Head                    -- ^ @h@
                 -> Term                    -- ^ @\@lab@
                 -> GrReader [(Label,Term)]
matchPattern_hlv h l1 =
    do
    tDtl <- asks termDirToLabels
    let listTsetL = Map.toAscList $ split_GE tDtl 
		    (TermDir (App h [l1, minTerm]))
	isGood (TermDir (App h' [l1',_])) = h' == h && l1' == l1
	isGood _ = False
	listTsetL' = takeWhile (isGood . fst) listTsetL
	expand (k,e) = [ (l,t) | let TermDir t = k , l <- Set.toList e ] 
    return $ concatMap expand listTsetL'

-- | Return all the matches for a @h(X,\@lab)@ pattern. Only arity 2.
matchPattern_hvl :: Head                    -- ^ @h@
                 -> Term                    -- ^ @\@lab@
                 -> GrReader [(Label,Term)]
matchPattern_hvl h l2 =
    do
    tItl <- asks termInvToLabels
    let listTsetL = Map.toAscList $ split_GE tItl
		    (TermInv (App h [minTerm,l2]))
	isGood (TermInv (App h' [_,l2'])) = h' == h && l2' == l2
	isGood _ = False
	listTsetL' = takeWhile (isGood . fst) listTsetL
	expand (k,e) = [ (l,t) | let TermInv t = k , l <- Set.toList e ] 
    return $ concatMap expand listTsetL'

-- *** Global Queries

-- | Return the size of the set of a 'Label'.
numberOfPlainTermsOfLabel :: Label -> GrReader Int
numberOfPlainTermsOfLabel l =
    do
    ltt <- asks labelToTerms
    return $ maybe 0 Set.size $ l `Map.lookup` ltt

-- | Return all the 'Label's that have a set in the Grammar.
knownLabels :: GrReader [Label]
knownLabels = asks $ Map.keys . labelToTerms

-- | Return the whole 'Grammar' as a list.
allKnownInfo :: GrReader [(Label,Term)]
allKnownInfo =
    do
    ltt <- asks labelToTerms
    let expand (k,e) =  [ (k,t) | t <- Set.toList e ]
    return $ concatMap expand $ Map.toList ltt

-- * Special Operations

-- | Join two 'Label's.
joinLabels :: Label    -- ^ The 'Label' that will hold the union set.
           -> Label    -- ^ The 'Label' that will disappear from the grammar.
           -> GrState ()
joinLabels l1 l2 | l1 == l2 = return ()
joinLabels l1 l2 =
    do
    gr <- get
    let set1, set2 :: Set.Set Term
        set1 =  fromMaybe Set.empty $ l1 `Map.lookup` labelToTerms gr
	set2 =  fromMaybe Set.empty $ l2 `Map.lookup` labelToTerms gr
        substLab :: Label -> Label
	substLab l | l==l2 = l1
	substLab l = l
        subst :: Term -> Term
	subst t@(Var _) = t
	subst (Lab l) = Lab $ substLab l
	subst (App h ts) = App h (Prelude.map subst ts)
        substPair :: (Label, Label) -> (Label, Label)
	substPair (la,lb) = (substLab la, substLab lb)
        joinedSet :: Set.Set Term
	joinedSet = Set.map subst (set1 `Set.union` set2)
        newLabToTerms :: Map.Map Label (Set.Set Term)
	newLabToTerms = Map.map (Set.map subst) $
			Map.insert l1 joinedSet $ 
                               l2 `Map.delete` labelToTerms gr
	newLabToTermsList = Map.toList newLabToTerms
	mkDirToLab (l,set) = [(TermDir t,Set.singleton l)| t <- Set.toList set]
        newDirToLab :: Map.Map TermDir (Set.Set Label)
	newDirToLab = Map.fromListWith Set.union $
		      concatMap mkDirToLab newLabToTermsList
	mkInvToLab (l,set) = [(TermInv t,Set.singleton l)| t <- Set.toList set]
        newInvToLab :: Map.Map TermInv (Set.Set Label)
	newInvToLab = Map.fromListWith Set.union $
		      concatMap mkInvToLab newLabToTermsList
    put $ gr{ labelToTerms    = newLabToTerms
	    , termDirToLabels = newDirToLab
	    , termInvToLabels = newInvToLab
	    , unknownIntersection = Set.map substPair $ 
                                    unknownIntersection gr
            , changed = True
            , joined = (l1,l2) : joined gr
	    }

-- | Check whether two sets are equal.
samePlainTermsOfLabels :: Label -> Label -> GrReader Bool
samePlainTermsOfLabels l1 l2 =
    do
    ltt <- asks labelToTerms
    let set1, set2 :: Maybe (Set.Set Term)
	set1 = l1 `Map.lookup` ltt 
	set2 = l2 `Map.lookup` ltt
    return $ set1 == set2


-- | Check whether two /plain/ terms intersect.
intersectingPlainTerms :: Bool -> Term -> Term -> GrReader Bool
-- ASSUMPTION : closed under transitivity.
intersectingPlainTerms emptyMatters term1 term2 =
    do
    let pair :: Term -> Term -> Term
        pair t1 t2 = App intersectionHead [t1,t2]
    let eq :: Term -> Term -> GrReader Bool
        eq (App h1 ts1) (App h2 ts2) | h1 == h2 =
              -- Ensured by type checking: length ts1 == length ts2 .
              mand [ plainElem intersectionLabel (pair t1 t2)
                   | (t1,t2) <- zip ts1 ts2 ]
        eq _ _ | not emptyMatters = return False
        eq (Lab l1) (Lab l2) = return $ l1 == l2
        eq _ _ = return False
    eq term1 term2

-- | Check whether the sets of two 'Label' intersect.
intersectingLabels :: Bool -> Label -> Label -> GrReader Bool
intersectingLabels emptyMatters lab1 lab2 =
    do
    ts1 <- plainTermsOfLabel lab1
    ts2 <- plainTermsOfLabel lab2
    mor [ intersectingPlainTerms emptyMatters t1 t2 | t1 <- ts1 , t2 <- ts2 ]

-- | Add the given pair to 'unknownIntersection'.
addToUnknownIntersection :: (Label,Label) -> GrState ()
addToUnknownIntersection p =
    do
    uI <- gets unknownIntersection
    modify $ \gr -> gr{unknownIntersection = p `Set.insert` uI }

-- | Update the 'unknownIntersection' and the set of 'intersectionLabel'.
computeIntersection :: Bool       -- ^ whether empty matters
		    -> GrState ()
computeIntersection emptyMatters =
    do
    { let pair :: Term -> Term -> Term
          pair t1 t2 = App intersectionHead [t1,t2]
    ; let checkPair :: (Label, Label) -> GrState ()
          checkPair p@(l1,l2) =
              do
              res <- actGr $ intersectingLabels emptyMatters l1 l2
              if res
                 then do
                      addToGr intersectionLabel $ pair (Lab l1) (Lab l2)
                      return ()
                 else addToUnknownIntersection p
    ; pairsToCheck <- gets $ Set.toList . unknownIntersection
    ; modify $ \gr -> gr{unknownIntersection = Set.empty}
    ; mapM_ checkPair pairsToCheck
    }

-- | Close the sets under the transitivity rule, that is: (Warshall's variant)
--
-- > | L1 : X , L2 : L1 => L2 : X .
--
--   Also return whether the 'Grammar' changed.
closeUnderTransitivity :: GrState ()
closeUnderTransitivity =
    do
    let add :: Label -> [Term] -> GrState ()
	add l = mapM_ (addToGr l)
	doOneLabel :: Label -> GrState ()
	doOneLabel l =
            do
            ts <- actGr $ plainTermsOfLabel l
            ls <- actGr $ labelsOfPlainTerm (Lab l)
	    mapM_ (\l' -> add l' ts) ls
    labels <- actGr knownLabels
    mapM_ doOneLabel labels

-- | Intersect two sets according to the rule schema
--
-- > | L1 : f(X1,...) , L2 : f(X2,...) , @!Inters : cons(X1,X2) 
-- >      => L : f(X1,...)
--
-- for each constructor @f@.
closeUnderIntersection :: Bool -> Intersection -> GrState ()
closeUnderIntersection emptyMatters (Intersection lab lab1 lab2) =
    {- For now, this is monotone, i.e. it only adds to lab. -}
    do
    let pair :: Term -> Term -> Term
        pair t1 t2 = App intersectionHead [t1,t2]
    let eq :: Term -> Term -> GrReader Bool
        eq (App h1 ts1) (App h2 ts2) | h1 == h2 =
            -- Ensured by type checking: length ts1 == length ts2 .
            mand [ plainElem intersectionLabel (pair t1 t2)
                 | (t1,t2) <- zip ts1 ts2 ]
        eq _ _ | not emptyMatters = return False
        eq (Lab l1) (Lab l2) = return $ l1 == l2
        eq _ _ = return False
    let doPair :: (Term,Term) -> GrState ()
	doPair (t1, t2) =
	    do
	    res <- actGr $ eq t1 t2
	    if res
	       then addToGr lab t1 >> return ()
	       else return ()
    ts1 <- actGr $ plainTermsOfLabel lab1
    ts2 <- actGr $ plainTermsOfLabel lab2
    mapM_ doPair $ [ (t1,t2) | t1 <- ts1, t2 <-  ts2]


-- * Pretty Printing

-- | Pretty print a 'Grammar'.
pprGr :: Renamer -> GrReader String
pprGr r =
    do
    labs <- knownLabels
    let pprSet []     = ""
	pprSet [t]    = pprTerm r t
	pprSet (t:ts) = pprTerm r t ++ ", " ++ pprSet ts
    let pprLab :: Label -> GrReader String
        pprLab l = 
            do
            ts <- plainTermsOfLabel l
            return $ realLabel r l ++ " : " ++ pprSet ts ++ " .\n"
    let cmpL l1 l2 = compare (realLabel r l1) (realLabel r l2)
    strs <- mapM pprLab $ sortBy cmpL labs
    let pprJoin (l1,l2) = "# " ++ realLabel r l2 ++ " <- " ++ 
                          realLabel r l1 ++ "\n"
    js <- asks joined
    let joins = if not (null js)
                then "# Joined states:\n" ++ concatMap pprJoin js 
                else ""
    return $ concat strs ++ joins

-- | Pretty print a 'Grammar' in order to save it in a file. Prepend a 
--   @\@@ to every @\'\$\'@ and @\'%\'@ so that it is parsable again. 
--   Output is sorted in order to behave deterministically.
pprSaveGr :: Renamer 
          -> Bool        -- ^ whether to print Join info as well
          -> GrReader String
pprSaveGr r printJoin =
    do
    labs <- knownLabels
    -- We sort _escaped_ lists, so hopefully fix(fix(x)) = fix(x),
    -- provided the heuristics are not confused by the escaping.
    let cmpL l1 l2 = compare (escapeLabel r l1) (escapeLabel r l2)
    let cmpT t1 t2 = compare (escapeTerm r t1) (escapeTerm r t2)
    let pprSet = concat . intersperse ", " . map (pprSaveTerm r)
    let pprLab :: Label -> GrReader String
	pprLab l =
            do
            ts <- plainTermsOfLabel l
            let Syntax.Label l' = escapeLabel r l
            return $ l' ++ " : " ++ pprSet (sortBy cmpT ts) ++ " .\n"
    strs <- mapM pprLab $ sortBy cmpL labs
    let pprJoin (l1,l2) = let Syntax.Label l1' = escapeLabel r l1
                              Syntax.Label l2' = escapeLabel r l2
                              in "# " ++ l2' ++ " <- " ++ l1' ++ "\n"
    js <- asks joined
    let joins = if printJoin 
                then "\n# Joined states:\n" ++ concatMap pprJoin js 
                else ""
    return $ concat strs ++ joins

-- * Utilities

-- | Augment a set in a map.
augmentSet :: (Ord k, Ord e) 
              => k    -- ^ The key of the 'Set'.
              -> e    -- ^ The element to add to the 'Set'.
              -> Map.Map k (Set.Set e) -> Map.Map k (Set.Set e)
augmentSet k e m =
    let set = e `Set.insert` (fromMaybe Set.empty $ k `Map.lookup` m)
	in set `seq` Map.insert k set m

-- | Restricts a 'Data.Map.Map' to keys greater than or equal to a given key.
--   (Similar to 'Data.Map.split')
split_GE :: Ord k => Map.Map k a -> k -> Map.Map k a
split_GE m key =
    let (_,point,post) = Map.splitLookup key m 
	in maybe post (\x -> Map.insert key x post) point

clampSet :: Ord e => e -> e -> Set.Set e -> Set.Set e
clampSet from to set =
    let (_,test,gt) = Set.splitMember from set
	ge = if test then Set.insert from gt else gt
	(lt,_) = Set.split to ge
	in lt
