{-# OPTIONS -cpp -fglasgow-exts #-}
-- #ignore-exports
module Main (main) where

import Data.Maybe
import Data.List hiding (union)
import Control.Monad.Reader
import Control.Monad.State
import System
import System.IO
import System.Time
import Char
import Text.ParserCombinators.Parsec (parse)

import qualified Syntax as S
import Renamer
import Parser
import ChomskyNormalForm
import Grammar
import TypeCheck
import Utility

-- * Main Monads

-- | The state used in the computation of the fixed point.
data St = St 
    { -- | The current term grammar.
      gr :: ! Grammar
      -- | The 'Renamer'.
    , renamer :: ! Renamer
      -- | The last generated label.
    , lastLabel :: ! Label
      -- | Index for the next label to be generated.
    , labelCount :: ! Int
      -- | Number of user specified labels in the .rules file.
    , userLabs :: ! Int
      -- | Number of generated labels, minus the number of joined labels.
    , newLabels :: ! Int
      -- | Bound on 'newLabels'. Constant set by user.
    , maxLabels :: ! Int
      -- | Number of joined labels.
    , joinCount :: ! Int
      -- | Bound on 'joinCount'. Constant set by user.
    , maxJoin   :: ! Int
      -- | Current stage in the computation of the fixed point.
    , stage :: ! Int
      -- | Whether the analysis failed.
    , failure :: ! Bool
      -- | Tells whether we must avoid interacting with the user. 
      --   Set by the user.
    , batchOpt :: ! Bool
      -- | Verbosity option set by user.
    , quietOpt :: ! Bool
      -- | The name of the @.rules@ file.
    , inputFile :: ! (Maybe String)
      -- | The name of the output file, if any.
    , outputFile :: ! (Maybe String)
      -- | Whether the output file should record the name of the input file 
      --   inside. Set by the user.
    , anonymousOpt :: ! Bool
      -- | Whether to put the rewriting rules into the output file, in spite of
      --   the output grammar begin closed under such rules.
    , outputRulesOpt :: ! Bool
      -- | Whether to allow for rewriting /ground/ terms which can not be
      --   expanded into a /pure/ term, but denote \"empty\" term sets.
    , emptyMattersOpt :: ! Bool
      -- | When rewriting @T@ for @\@l:T@, whether to prefer to use @\@l@ 
      --   in the rewritten @T@, if possible. Used by 'addPlainTermSomewhere'.
    , useHintsOpt :: ! Bool
      -- | Whether to prefer user @\@labels@ rather than @\$labels@.
      --   Used by 'addPlainTermSomewhere'.
    , preferUserLabelsOpt :: ! Bool
    , joinUserLabelsOpt :: ! Bool
      -- | The normalized rules we are closing under.
    , rules :: ! [Rule]
      -- | The (not yet normalized) rules from the input file.
      --   We store them here to eventually copy them into the output file.
    , rawRules :: ! [S.Rule]
      -- | The intersection rules.
    , intersections :: ! [Intersection]
    }

-- | Read-only (w.r.t. 'St') non-deterministic computation.
type NonDetReader = ReaderT St []

-- | Read-write (w.r.t. 'St') deterministic IO-enabled computation.
type IOState = StateT St IO

-- * Merging Grammar 'Monad's into Ours

-- | Perform a read-write actions on the 'Grammar'.
doGr :: GrState a -> IOState a
doGr action = 
    do
    (res,newGr) <- gets $ runState action . gr
    modify (\st -> st{gr = newGr})
    return res

-- | Perform a /deterministic/ 'Grammar' read.
readGr :: GrReader a -> IOState a
readGr action = gets $ runReader action . gr

-- | Perform a /non-deterministic/ 'Grammar' read.
pickFromGr :: GrReader [a] -> NonDetReader a
pickFromGr action = asks gr >>= lift . runReader action

-- * Pattern Matching

-- | Match a 'Term' pattern with a /ground/ 'Term', non-deterministically 
--   returning the unifying 'Subst's.

-- ASSUMPTION Sets are closed by transitivity.
matchTermWithGroundTerm :: Term -> Term -> NonDetReader Subst
matchTermWithGroundTerm (Var v) t = 
    do
    let readFromGr :: GrReader a -> NonDetReader a
	readFromGr action = asks gr >>= return . runReader action
    emptyMatters <- asks emptyMattersOpt
    empty <- readFromGr (isEmptyPlainTerm t)
    if empty && not emptyMatters
       then lift [] -- match failure
       else return $ Subst [(v,t)]
matchTermWithGroundTerm (Lab l) (Lab l') = 
    if l == l'
       then return idSubst
       else lift []
matchTermWithGroundTerm (App h1 ts1) (App h2 ts2) =
    do
    guard $ h1 == h2
    -- guard $ length ts1 == length ts2  Statically checked.
    res <- zipWithM matchTermWithGroundTerm ts1 ts2
    return $ mergeSubsts res
matchTermWithGroundTerm _ _ = lift [] -- Match failure.

-- | Match a /plain/ 'Term' pattern somewhere in the grammar.
matchSomewhere_c :: Variable -> Term -> NonDetReader Subst
matchSomewhere_c v term =
    do
    l <- pickFromGr $ labelsOfPlainTerm term
    return $ Subst [(v,Lab l)]

-- | Match a 'Term' pattern of the form @h(X)@ somewhere in the grammar.
matchSomewhere_hv :: Variable -> Head -> Variable ->
                    NonDetReader Subst
matchSomewhere_hv v h x =
    do
    (l,(App _ [t_x])) <- pickFromGr $ matchPattern_h h
    return $ Subst [(v,Lab l),(x,t_x)]

-- | Match a 'Term' pattern of the form @h(X,Y)@ somewhere in the grammar.
matchSomewhere_hvv :: Variable -> Head -> Variable -> Variable ->
                      NonDetReader Subst
matchSomewhere_hvv v h x y =
    do
    (l,(App _ [t_x,t_y])) <- pickFromGr $ matchPattern_h h
    return $ Subst [(v,Lab l),(x,t_x),(y,t_y)]

-- | Match a 'Term' pattern of the form @h(\@lab,Y)@ somewhere in the grammar.
matchSomewhere_hlv :: Variable -> Head -> Term -> Variable ->
                      NonDetReader Subst
matchSomewhere_hlv v h t_x y =
    do
    (l,(App _ [_,t_y])) <- pickFromGr $ matchPattern_hlv h t_x
    return $ Subst [(v,Lab l),(y,t_y)]

-- | Match a 'Term' pattern of the form @h(X,\@lab)@ somewhere in the grammar.
matchSomewhere_hvl :: Variable -> Head -> Variable -> Term ->
                      NonDetReader Subst
matchSomewhere_hvl v h x t_y =
    do
    (l,(App _ [t_x,_])) <- pickFromGr $ matchPattern_hvl h t_y
    return $ Subst [(v,Lab l),(x,t_x)]

-- | Generic pattern matching for a 'Clause'. Yields all matching 
--   substitutions, non-deterministically.
matchClause :: Clause -> NonDetReader Subst
matchClause (Clause (LLab lab) (App h ts1))= 
    do
    term' <- pickFromGr $ plainTermsOfLabel_h lab h
    let (App _ ts2) = term'
    res <- zipWithM matchTermWithGroundTerm ts1 ts2
    return $ mergeSubsts res
matchClause (Clause (LLab lab) term)= 
    do
    term' <- pickFromGr $ plainTermsOfLabel lab
    matchTermWithGroundTerm term term'
matchClause (Clause (LVar v) term@(App _ [])) =
    matchSomewhere_c v term
matchClause (Clause (LVar v) (App h [Var v'])) = 
    matchSomewhere_hv v h v'
matchClause (Clause (LVar v) term@(App _ [Lab _])) = 
    matchSomewhere_c v term
matchClause (Clause (LVar v) term@(App _ [Lab _, Lab _])) = 
    matchSomewhere_c v term
matchClause (Clause (LVar v) (App h [l@(Lab _), Var v'])) =
    matchSomewhere_hlv v h l v'
matchClause (Clause (LVar v) (App h [Var v', l@(Lab _)])) =
    matchSomewhere_hvl v h v' l
matchClause (Clause (LVar v) (App h [Var v', Var v''])) =
    matchSomewhere_hvv v h v' v''
matchClause (Clause (LVar v) term) =
    do -- Scan everything (inefficient, but seldom necessary).
    (l,t) <- pickFromGr $ allKnownInfo
    let sLoc = Subst [(v,Lab l)]
    subst <- matchTermWithGroundTerm (sLoc `applySubstTerm` term) t
    return $ sLoc `mergeSubst` subst

matchRule :: Rule -> NonDetReader (Label, Term)
matchRule ([] :=> (Clause (LLab lab) term)) =
    return (lab, term)
matchRule ((cl:cls) :=> goal) =
    do
    subst <- matchClause cl
    let newCls  = map (applySubst subst) cls
        newGoal = subst `applySubst` goal   
    matchRule (newCls :=> newGoal)
matchRule _ =
    error "matchRule: internal error"

-- | Apply all 'Rule's and return (non-deterministically) all the /ground/
--   'Terms' that would need to be added to 'Label's.
matchRules :: [Rule] -> NonDetReader (Label, Term)
matchRules rs =
    do
    r <- lift $ rs
    matchRule r

-- * Stage Driver

-- | Check whether we can add a fresh 'Label' to 'gr'.
mayCreateLabels :: IOState Bool
mayCreateLabels = do st <- get
		     let availableLabs = newLabels st < maxLabels st
		     return $ (stage st >= 3) && availableLabs

-- | Check whether the join bound has been reached.
mayJoinLabels :: IOState Bool
mayJoinLabels = do st <- get
		   return $ maxJoin st == -1 || joinCount st < maxJoin st

-- | Check whether we can ask the user at this stage.
mayPromptUser :: IOState Bool
mayPromptUser =  
    do 
    batch    <- gets batchOpt
    curStage <- gets stage
    if batch
       then return False
       else return $ curStage >= 3

-- | Check whether we already tried all the strategies, including user 
--   interaction. In this case, we need to reuse a 'Label', hoping not to
--   lose too much precision in the approximation.
triedEverything :: IOState Bool
triedEverything = do
		  curStage <- gets stage
		  return $ curStage >= 3

-- | Check whether the analisys has failed.
alreadyFailed :: IOState Bool
alreadyFailed = 
    do
    failed <- readGr $ plainElem resultLabel (App resultHead [])
    when failed $ modify (\st -> st{failure = True})
    return failed

-- | Insert a new 'Term' into the 'Grammar', incrementing the 'Int' if not
--   possible now.
handleMatch :: Int -> (Label, Term) -> IOState Int
handleMatch notHandled (lab,term) =
    do res <- addGroundTermToSet lab term
       return $ maybe (succ notHandled) (const notHandled) res

-- | Main driver for all 'stage's. Loops until the state reaches a fixed point.
closeUnderRules :: IOState ()
closeUnderRules =
    do
    curStage <- gets stage
    when (curStage > 1)
	     (do
	      modify (\st -> st{stage = curStage - 1})
	      closeUnderRules
	      modify (\st -> st{stage = curStage})
	     )
    doGr clearGr
    quiet <- gets quietOpt
    unless quiet (do displayGrammar ; displayUsedLabels)
    lift (do
          unless quiet $ putStrLn ""
	  putStr $ "***** stage " ++ show curStage
	  hFlush stdout
	 )
    rs <- gets rules
    let computeMatches :: IOState [(Label, Term)]
	computeMatches = gets $ runReaderT (matchRules rs)
    case curStage of
	{ 1 -> do res <- mayJoinLabels 
	          when res tryToJoinLabels
	; 2 -> do matches <- computeMatches
	          toDo <- foldM handleMatch 0 matches 
                  lift $ putStr $ " (" ++ show toDo ++ " to do)"
	          doGr closeUnderTransitivity
	          emptyMatters <- gets emptyMattersOpt
	          inters <- gets intersections
	          doGr $ mapM_ (closeUnderIntersection emptyMatters) inters
	          doGr $ computeIntersection emptyMatters
	; 3 -> do matches <- computeMatches
                  mor [ do
			uncurry addGroundTermToSet match
			gets $ changed . gr              |
			match <- matches ]
	          return ()
	; _ -> error "closeUnderRules: unknown stage"
	}
    lift $ putChar '\n'
    wasChanged <- gets $ changed . gr
    failed <- alreadyFailed
    when (wasChanged && not failed) closeUnderRules

-- * Grammar Changes

-- | Return a fresh 'Label'.
newLabel :: IOState Label
newLabel =
    do
    st <- get
    let newL = Label (labelCount st)
    put $ st{ lastLabel  = newL
	    , labelCount = succ $ labelCount st
	    , newLabels  = succ $ newLabels st 
	    }
    let r = renamer st
    lift $ putStr $ "\n##### Created state " ++ realLabel r newL ++ " -- "
    displayUsedLabels
    return newL

-- | Add a /plain/ 'Term' to the set of 'Label' in the 'gr'.
--   Also, print a dot for every added 'Term'.
addPlainTermToSet :: Label -> Term -> IOState ()
addPlainTermToSet lab term =
    do
    alreadyThere <- doGr $ addToGr lab term
    unless alreadyThere $
                   lift $ putChar '.' >> hFlush stdout

-- | Try to add a /ground/ 'Term' to the set of 'Label' in the 'gr'.
--   If necessary, the production is changed into /Chomsky Normal Form/.
addGroundTermToSet :: Label -> Term -> IOState (Maybe ())
addGroundTermToSet lab term@(Lab _) = 
    addPlainTermToSet lab term >> return (Just ())
addGroundTermToSet lab (App h ts) =
    do
    subRes <- mapM (addGroundTermSomewhere $ Just lab) ts
    if all isJust subRes
       then do
	    let labels = map (Lab . fromJust) subRes
	    addPlainTermToSet lab (App h labels)
	    return $ Just ()
       else return Nothing
addGroundTermToSet _ _ = error "addGroundTermToSet"

-- | Try to find a place for a /plain/ 'Term' anywhere in the 'gr', 
--   choosing the \'best\' 'Label'.  If possible, do not change 'gr'.
addGroundTermSomewhere :: Maybe Label      -- ^ Hint: try this first.
		       -> Term
		       -> IOState (Maybe Label)
addGroundTermSomewhere _ (Lab l) = return $ Just l
addGroundTermSomewhere hint (App h ts) =
    do
    subRes <- mapM (addGroundTermSomewhere hint) ts
    if all isJust subRes
       then do
	    let labels = map (Lab . fromJust) subRes
	    addPlainTermSomewhere hint (App h labels)
       else return Nothing
addGroundTermSomewhere _ _ = error "addGroundTermSomewhere"

-- | Try to find a place for a /plain/ 'Term' anywhere in the 'gr', 
--   choosing the \'best\' 'Label'.  If possible, do not change 'gr'.
--   On failure (e.g., early stage, so we want to try the alternatives first), 
--   return 'Nothing'; otherwise return 'Just' 'Label'.
addPlainTermSomewhere :: Maybe Label  -- ^ Hint: try this first.
                      -> Term
                      -> IOState (Maybe Label)
addPlainTermSomewhere _ (Lab l) = 
    return $ Just l
addPlainTermSomewhere (Just hintLabel) term =
    do
    let noHint = addPlainTermSomewhere Nothing term
    let hint = return $ Just hintLabel
    mif (gets useHintsOpt)
        (mif (readGr $ plainElem hintLabel term) hint noHint)
        noHint
addPlainTermSomewhere _ term =
    do
    -- Heuristic : find a suitable label for term.
    st <- get
    labs <- readGr $ labelsOfPlainTerm term
    case labs of
        l@(l1:_) -> -- term occurs in the grammar.
	    do -- Choose the 'best' l.
            let isUser :: Label -> Bool
                isUser (Label n) = n >= reservedLabels && n < userLabs st
		userLab :: Label
		userLab = case filter isUser l of
			      []     -> l1
			      (l2:_) -> l2
	    liftM Just $ mif' (gets preferUserLabelsOpt) userLab l1
	[] -> -- term does not occur in the grammar.
	    do
	    { let strategies :: [(IOState Bool, IOState (Maybe Label))]
	          strategies = 
	             [(mayCreateLabels,
		       do
		       lab <- newLabel
		       addPlainTermToSet lab term
		       return $ Just lab
		      ),
		      (mayPromptUser,
		       promptUser term
		      ),
		      (triedEverything,
		       -- Fallback strategy: add to the last label.
		       do 
		       r <- gets renamer
		       lab <- gets lastLabel
		       lift $ putStr
	                 ("\n##### Fallback " ++ realLabel r lab ++ " : " ++ 
			  pprTerm r term)
		       addPlainTermToSet lab term
		       return $ Just lab
		      )
		     ]
	    ; let tryStrategies :: [(IOState Bool, IOState (Maybe Label))] 
	                        -> IOState (Maybe Label)
	          tryStrategies ((chk,str):rest) = 
	              do 
	              { c <- chk
		      ; if c 
			then do { res <- str 
				; if isJust res 
				  then return res
				  else tryStrategies rest
				}
			else tryStrategies rest
		      }
	          tryStrategies [] = return Nothing
	    ; tryStrategies strategies
	    }

{- | 
   Try to reclain some 'Label' slots by joining two 'Labels' if they have
   similar sets.

   [NOTE] In general, joining labels might lead to a non terminating 
   algorithm, since joining makes the grammar size to decrease. Currently, 
   we enforce termination by not allowing more than 'maxJoin' joins.
-}
tryToJoinLabels :: IOState ()
tryToJoinLabels =
    do
    joinUser <- gets joinUserLabelsOpt
    labs <- gets userLabs
    rs <- gets rules
    is <- gets intersections
    let -- As 'readGr', but returns 'NonDetReader' instead of 'IOState'.
        -- Sadly, we can not generalize this since being a MonadState does not
        -- imply being a MonadReader.
        readGr' :: GrReader a -> NonDetReader a
        readGr' action = asks $ runReader action . gr

        joinable :: NonDetReader (Label,Label)
        joinable =
            do
            l1 <- pickFromGr knownLabels
            l2 <- pickFromGr knownLabels
            guard $ l1 /= l2
            let Label n = l2
            if joinUser
               then do guard $ n >= reservedLabels
                       guard $ not $ l2 `occursInRules` rs
                       guard $ not $ l2 `occursInIntersections` is
               else do guard $ n > labs
            guard =<< readGr' (samePlainTermsOfLabels l1 l2)
            return (l1,l2)
    joinList <- gets $ runReaderT joinable
    case joinList of
	[] -> return ()
	((l1, l2):_) -> 
	    do 
	    let subst l = if l == l2 then l1 else l
	    doGr $ joinLabels l1 l2
	    modify $ \st -> st{ newLabels = pred $ newLabels st
			      , lastLabel = subst $ lastLabel st 
			      , joinCount = succ $ joinCount st
			      }
	    count <- gets joinCount
	    r <- gets renamer
	    lift $ putStrLn $ 
		     "\n##### Joined states " ++ realLabel r l1 ++ " <- " ++ 
		     realLabel r l2 ++ "        (Join #" ++ show count ++ ")"
	    res <- mayJoinLabels
	    unless res $ lift $ putStrLn $
		     "\n##### WARNING: maxJoin bound reached !"
	    displayUsedLabels

-- * User Interaction

-- | Print the 'gr' to the user.
displayGrammar :: IOState ()
displayGrammar =
    do
    ren <- gets renamer
    str <- readGr $ pprGr ren
    lift $ putStr str

-- | Print information about available 'Label' slots to the user.
displayUsedLabels :: IOState ()
displayUsedLabels =
    do
    st <- get 
    lift $ putStrLn ("Used " ++ (show $ newLabels st) ++ 
		     " of "  ++ (show $ maxLabels st) ++ 
		     " states")

displayStatistics :: IOState ()
displayStatistics =
    do
    prods <- readGr $ allKnownInfo
    let prodsNum = show $ length prods
    lift $ putStrLn $ "Automaton has " ++ prodsNum ++ " transitions."

-- | Ask the user what to do, i.e. ask for a 'Label' whose set is where to 
--   place the 'Term'.
--
--   If 'Term' is 'MinTerm', do not ask anything, but let the user explore the
--   state.
promptUser :: Term -> IOState (Maybe Label)
promptUser termToPlace = 
    let showState :: IOState ()
        showState =
	    do
	    st <- get
	    lift $ putStr $ unlines 
		  [ "lastLabel  " ++ show (lastLabel st)
		  , "labelCount " ++ show (labelCount st)
		  , "newLabels  " ++ show (newLabels st)
		  , "maxLabels  " ++ show (maxLabels st)
		  , "stage      " ++ show (stage st)
		  , "changed    " ++ show (changed $ gr st)
		  ]
        -- | Show /pure/ term intances of a /plain/ term.
        showPureExamples :: Term -> IOState ()
	showPureExamples term =
	    do
	    r <- gets renamer
	    let showTerms :: Int -> [Term] -> IO ()
	        showTerms _ [] = putStrLn "<end>"
		showTerms 0 _  = putStrLn "..."
		showTerms _ [t] = putStrLn (pprTerm r t ++ "\n<end>") 
		showTerms n (t:ts) = 
		    do 
		    putStrLn (pprTerm r t ++ ",")
		    showTerms (n-1) ts
	    st <- get
	    let examples = runReaderT (pureExamples [] term) st
	    lift $ showTerms 5 examples
        -- | Show /pure/ term instances of a /plain/ given term, or of 
        --   termToPlace if no term is given
        showPureExamplesFromString :: String -> IOState ()
        showPureExamplesFromString s =
            case dropWhile isSpace s of
                ""  -> showPureExamples termToPlace
                str -> case parse plainTerm "-" str of
                           Left pe    -> lift $ print pe
                           Right term -> do r <- gets renamer
                                            case renameTerm r term of
                                             Nothing -> lift $
                                                        putStrLn "Unknown term"
                                             Just t  -> showPureExamples t
        -- | Check whether a /pure/ term belongs to a set.
        checkTerm :: String -> IOState ()
	checkTerm str =
	    if not (' ' `elem` str)
	       then lift $ putStrLn "Syntax error: t @a T"
	       else do 
	       { r <- gets renamer
               ; let (l,_:t) = break (isSpace) str
		     lab = renameLabel' r l
               ; if lab == Nothing
                 then lift $ putStrLn "Unknown state"
                 else do
               { let lab' = fromJust lab
               ; known <- readGr $ knowsLabel lab'
	       ; if not known
		 then lift $ putStrLn "Error: state does not exist"
		 else 
		 case parse pureTerm "-" t of
		 Left pe    -> lift $ print pe
		 Right term -> 
                     case renameTerm r term of
                       Nothing -> lift $ putStrLn "Unknown term"
                       Just term' -> 
                           do
                           res <- readGr $ pureElem lab' term'
                           lift $ putStrLn $
		                if res
		                then "@@ Term belongs to the set."
		                else "@@ Term does NOT belong to the set."
	       }}
        showHelp :: IOState ()
        showHelp = lift $ putStr $ unlines
            [ " Available commands:"
            , "  s g    : show automaton"
            , "  s t    : show the term we are trying to place"
            , "  s x    : show pure instances of that term"
            , "  s x T  : show pure instances of a given plain term"
            , "  l @a   : assign state @a to term"
            , "  l      : don't assign a state for the term now"
            , "  s s    : show statistics"
            , "  t @a T : test whether the pure term T belongs to [@a]"
            , "  q      : quit"
            ]
        -- | The main user interaction loop
	loop :: IOState (Maybe Label)
	loop =
	    do
	    lift $ putStr "|> " >> hFlush stdout
	    l <- lift $ getLine
	    r <- gets renamer
	    case l of
	        "s g" -> do displayGrammar ; displayUsedLabels ; loop
		"s t" -> do lift (putStrLn $ pprTerm r termToPlace) ; loop
		('s':' ':'x':rest) -> do showPureExamplesFromString rest ; loop
		"s s" -> do showState ; loop
		"l"   -> return Nothing
		"q"   -> lift $ exitWith $ ExitSuccess
		('l':' ':rest) -> 
                         case renameLabel' r rest of
                           Nothing  -> do lift $ putStrLn "Unknown state"
                                          loop
                           Just lab -> do addGroundTermToSet lab termToPlace
				          return $ Just lab
		('t':' ':rest) -> do checkTerm rest ; loop
                "help" -> do showHelp ; loop
                "?"    -> do showHelp ; loop
		_      -> do lift $ putStrLn "?? Unknown command (try 'help')"
                             loop
    in do
       r <- gets renamer
       when (termToPlace /= minTerm)
	    (lift $ putStrLn $ 
	     "\n** I need a state for term " ++ pprTerm r termToPlace)
       loop

-- | Instantiate a /ground/ 'Term' into a /pure/ 'Term' in all the possible
--   ways, in a non-deterministic fashion. Avoid infinite loops on the grammar
--   reachability graph.
pureExamples :: [Label] -> Term -> NonDetReader Term
pureExamples visited (App h ts) = 
    do 
    groundTs <- mapM (pureExamples visited) ts
    return $ App h groundTs
pureExamples visited (Lab l) =
    if l `elem` visited
       then lift []
       else do
            t <- pickFromGr $ plainTermsOfLabel l
	    pureExamples (l:visited) t
pureExamples _ (Var _) = error "pureExamples: variable found"

-- | Parse and type check a @.rules@ file.  If succeeds, returns the rules
--   in Chomsky Normal Form, and stores the original ones in the 'St' state.
parseFile :: IOState Bool
parseFile =
    do
    fn <- gets $ fromJust . inputFile
    h <- lift $ openFile fn ReadMode
    text <- lift $ hGetContents h
    -- Parse all the file.
    case parse rulesFile fn text of
        Left parseError -> 
            do
            lift $ hPrint stderr parseError
            lift $ hClose h
	    return False
        Right (simpleRules, fullRules, inters, grammar) -> 
            do
            lift $ hClose h
	    -- Type check everything.
	    case typeCheckRulesAndGrammar simpleRules fullRules grammar of
	      Just errorValue -> 
		  do
		  lift $ putStrLn $ uncurry typeErrorMsg errorValue
		  return False
	      Nothing ->
		  do
		  let normalizedGrammar = normalizeProductions grammar
		      -- Normalize only the simple rules.
		      normalizedSimpleRules = normalizeSimpleRules simpleRules
		      normalizedRules = normalizedSimpleRules ++ fullRules
		      parsedRules = simpleRules ++ fullRules
		      -- Rename all the labels.
		      (ren, renamedGrammar, renamedRules, renamedInters, lab) =
			  rename normalizedGrammar normalizedRules inters
                  -- Fill the sets with the initial grammar.
		  let addRow :: (Label, [Term]) -> GrState ()
                      addRow (l,ts) = mapM_ (addToGr l) ts
                  doGr $ mapM_ addRow renamedGrammar
		  modify (\st -> st{ renamer = ren })
                  -- Save the original, non-normalized, rules.
                  modify (\st -> st{ rawRules = parsedRules })
		  -- Store the normalized rules.
                  modify (\st->
			  st{ rules = renamedRules
			    , intersections = renamedInters
			    , labelCount = lab
			    , userLabs = lab })
		  return True 

-- | Save the grammar in a file, if it was requested.
saveFixedPoint :: IOState ()
saveFixedPoint =
    do
    { outFn <- gets $ fromJust . outputFile
    ; inFn  <- gets $ fromJust . inputFile
    ; anon  <- gets anonymousOpt
    ; let inFnInside = if anon then "<<unknown>>" else inFn
    ; out <- lift $ openFile outFn WriteMode
    ; ren <- gets renamer
    ; strGr <- readGr $ pprSaveGr ren (not anon)
    ; saveRules <- gets outputRulesOpt
    ; strRules <- if not saveRules
                  then return
 "# The 'rule' part is intentionally empty, since this is a fixed point."
                  else do
                       rawR <- gets rawRules
                       return $ concatMap ((++ " .\n") . S.pprGeneralRule) rawR
    ; lift $ do 
 { hPutStr out $ unlines
   [ "######################################################################"
   , "# This file was automatically generated as the fixed point of the"
   , "# automaton and rules in the \"" ++ inFnInside ++ "\" file."
   , ""
   , "# The rewriting rules follow."
   , ""
   , strRules
   , ""
   , "%%"
   , ""
   , "# The automaton follows."
   , ""
   , strGr
   , ""
   , "# End of generated file."
   ]
 ; hClose out
 ; putStrLn $ "%%%%% Fixed point saved in \"" ++ outFn ++ "\"."
 } 
    }

-- | Parse command line options and update 'St' accordingly.
parseOptions :: [String] -> IOState ()
parseOptions ("-n" : opts') =
    case opts' of
        n : opts -> do modify (\st -> st{maxLabels = read n})
                       parseOptions opts
        _        -> usageError
parseOptions ("-j" : opts') =
    case opts' of
        j : opts -> do modify (\st -> st{maxJoin = read j})
                       parseOptions opts
        _        -> usageError
parseOptions ("-o" : opts') =
    case opts' of
        o : opts -> do modify (\st -> st{outputFile = Just o})
                       parseOptions opts
        _        -> usageError
parseOptions ("--quiet" : opts) =
    do modify (\st -> st{quietOpt = True}) ; parseOptions opts
parseOptions ("--batch" : opts) =
    do modify (\st -> st{batchOpt = True}) ; parseOptions opts
parseOptions ("--anonymous" : opts) =
    do modify (\st -> st{anonymousOpt = True}) ; parseOptions opts
parseOptions ("--output-rules" : opts) =
    do modify (\st -> st{outputRulesOpt = True}) ; parseOptions opts
parseOptions ("--empty-matters" : opts) =
    do modify (\st -> st{emptyMattersOpt = True}) ; parseOptions opts
parseOptions ("--no-hints" : opts) =
    do modify (\st -> st{useHintsOpt = False}) ; parseOptions opts
parseOptions ("--avoid-user" : opts) =
    do modify (\st -> st{preferUserLabelsOpt = False}) ; parseOptions opts
parseOptions ("--restrict-join" : opts) =
    do modify (\st -> st{joinUserLabelsOpt = False}) ; parseOptions opts
parseOptions ("--help" : _) = usageError
parseOptions [fn] = modify (\st -> st{inputFile = Just fn})
parseOptions _ = usageError

-- | Display usage.
usageError :: IOState ()
usageError =
    lift $ do
    { prog <- getProgName
    ; hPutStr stderr $ unlines
 [ "Usage: " ++ prog ++ " [options] input.rules"
 , "  Available options:"
 , "    -n maxStates    Use maxStates states for the analysis."
 , "                        [default " ++ show (maxLabels initialState) ++ "]"
 , "    -j maxJoin      Do not perform more than maxJoin Join operations."
 , "                        [default " ++ show (maxJoin initialState) ++ 
                                           "] (-1 means infinite)"
 , "    -o output.rules Save the fixed point in a file."
 , "    --quiet         Do not show the automaton at each stage."
 , "    --batch         Do not interact wth the user."
 , "    --anonymous     Do not store the name of the input .rules file"
 , "                    into a comment in the output file."
 , "    --output-rules  Put rewriting rules into the output file, even if"
 , "                    the output automaton is closed under such rules."
 , "    --empty-matters Allow for rewriting of ground terms with no pure"
 , "                    descendant."
 , "    --no-hints      Disregards hints in heuristics."
 , "    --avoid-user    Do not prefer user-defined states in heuristics."
 , "    --restrict-join Do not join user states."
 , "    --help          Display this help message."
 ]
    ; exitFailure
    }


initialState :: St
initialState = St{ gr                  = emptyGr
		 , renamer             = emptyRenamer
		 , stage               = 3                  -- The last stage
                 , failure             = False
		 , lastLabel           = Label (-100)       -- "$lastLabel"
		 , labelCount          = -100
		 , userLabs            = -100
		 , newLabels           = 0
		 , maxLabels           = 1
		 , joinCount           = 0
		 , maxJoin             = -1                 -- unlimited
		 , batchOpt            = False
		 , quietOpt            = False
		 , inputFile           = Nothing
                 , outputFile          = Nothing
                 , anonymousOpt        = False
                 , outputRulesOpt      = False
		 , emptyMattersOpt     = False
		 , useHintsOpt         = True
		 , preferUserLabelsOpt = True
                 , joinUserLabelsOpt   = True
		 , rules               = []
                 , rawRules            = []
		 , intersections       = []
		 }

main :: IO ()
main = 
    do 
    let displayFixedPoint :: IOState ()
        displayFixedPoint =
            do
	    lift $ putStr $ unlines
                     [ "%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%"
		     , ""
		     , "         Fixed point found !!!"
		     , ""
                     ]
	    displayGrammar
	    displayUsedLabels
	    displayStatistics
    let driver :: IOState ()
        driver = 
	    do
            -- Parse command line options.
	    args <- lift $ getArgs
	    parseOptions args
            -- Check that inputFile was specified.
            input <- gets inputFile
            if isNothing input
               then usageError
               else return ()
            -- Parse .rules file.
	    res <- parseFile
	    if not res
	       then lift $ exitFailure
	       else do
                    -- Find the minimum fixed point.
		    startTime <- lift $ getClockTime
		    closeUnderRules
		    endTime   <- lift $ getClockTime
                    -- Print the fixed point, if requested.
                    willSave <- gets $ isJust . outputFile
                    quiet    <- gets quietOpt
                    unless (quiet && willSave) displayFixedPoint
                    -- Print evaluation time.
	            let diff = timeDiffToString $ 
		               diffClockTimes endTime startTime
	                message = if diff == "" 
			          then "Done." 
			          else "Done (" ++ diff ++ ")."
	            lift $ putStrLn $ "%%%%% " ++ message
                    failed <- gets failure
                    when failed $ lift $ putStrLn $
                        "%%%%% Alas, analysis has !FAIL!ed !!!!"
                    -- Save the output in a file, if requested.
                    when willSave $ saveFixedPoint
                    -- Unless we are in batch mode, interact with the user.
		    batch <- gets batchOpt
		    unless batch $ promptUser minTerm >> return ()
                    when failed $ lift exitFailure
    evalStateT driver initialState
