{-# OPTIONS -cpp #-}
-- #ignore-exports
module Parser (rulesFile, plainTerm, pureTerm) where

import Prelude hiding (head)
import Syntax

import Text.ParserCombinators.Parsec hiding (label)
import Text.ParserCombinators.Parsec.Char 

data GeneralRule 
    = SimpleRule Rule  -- ^ #SimpleRule# 
                       --   Simple rewriting rule such as @f(X,Y) => f(Y,X)@.
    | FullRule   Rule  -- ^ #FullRule#
                       --   Complex rule, may use custom pattern matching.
    | IntersRule Intersection
                       -- ^ #IntersRule#
                       --   Intersection of two labels.

-- | One of some non alphanumeric symbols.
symbol :: Parser Char
symbol = oneOf "\'\";+-*/=^&_"

-- | @#-till-newline@ comment parser.
comment :: Parser ()
comment =
    do
    char '#' -- begin of comment
    many $ noneOf ['\n']
    char '\n'
    return ()

-- | Whitespace parser (comments are whitespace).
sp :: Parser ()
sp = 
    do 
    many ((space >> return ()) <|> comment)
    return ()

-- | Heads start with lowercase char.
head :: Parser Head
head =
    do
    u <- (lower <|> digit <|> symbol <|> char '!')
    us <- many (upper <|> lower <|> digit <|> symbol <|> oneOf "$!%_") 
    sp
    return $ Head $ u:us

-- | Labels start with @\@@. We let the user to fiddle with 
--   'intersectionLabel', since we allow @\"\@\!Inters\"@ to be parsed, but
--   use it with cure!
label :: Parser Label
label =
    do
    u <- char '@'
    us <- many (upper <|> lower <|> digit <|> symbol <|> oneOf "$!%_")
    sp
    return $ Label $ u:us

-- | 'Term' variant of 'label'.
labelTerm :: Parser Term
labelTerm = label >>= return . Lab

-- | Variables start with an uppercase char. 
--   @\'\!\'@ is reserved inside variables.
variable :: Parser Variable
variable =
    do
    u <- upper
    us <- many (upper <|> lower <|> digit <|> symbol <|> char '$')
    sp
    return $ Variable $ u:us

-- | 'Term' variant of 'variable'.
variableTerm :: Parser Term
variableTerm = variable >>= return . Var

-- | Either a 'labelTerm' or a 'variableTerm'.
labelOrVariableTerm :: Parser Term
labelOrVariableTerm = labelTerm <|> variableTerm

loc :: Parser Loc
loc = (do v <- variable ; return $ LVar v)
      <|> (do l <- label ; return $ LLab l)

-- | Parse only the top-level of a term, using the given parser to parse
--   inner terms.
termFactory :: Parser Term -> Parser Term
termFactory termCont =
    do 
    h <- head
    sp
    ((do
      { ts <- (between (char '(' >> sp) (char ')' >> sp)
	       (sepBy termCont (char ',' >> sp)))
      ; sp
      ; return $ App h ts
      }) <|> 
     (return $ App h []))

--  Parse /ground/ terms.
groundTerm :: Parser Term
groundTerm = labelTerm <|> termFactory groundTerm

-- | Parse /simple/ terms.
simpleTerm :: Parser Term
simpleTerm = variableTerm <|> termFactory simpleTerm

-- | Parse /pure/ terms.
pureTerm :: Parser Term
pureTerm = termFactory pureTerm

-- | Parse /plain/ terms.
plainTerm :: Parser Term
plainTerm = labelTerm <|> termFactory labelTerm

-- | Parse terms of depth <= 1.
flatTerm :: Parser Term
flatTerm = labelOrVariableTerm <|> termFactory labelOrVariableTerm

-- | Parse arbitrary terms.
fullTerm :: Parser Term
fullTerm = labelOrVariableTerm <|> termFactory fullTerm

flatClause :: Parser Clause
flatClause =
    do
    l <- loc
    char ':'
    sp
    t <- flatTerm
    return $ Clause l t

fullClause :: Parser Clause
fullClause =
    do
    l <- loc
    char ':'
    sp
    t <- fullTerm
    return $ Clause l t


-- | 'IntersRule's start with @'||'@. Example
--
-- > || @a : @b /\ @c .
intersRule :: Parser GeneralRule
intersRule =
    do
    char '|'
    sp
    lab <- label
    char ':'
    sp
    lab1 <- label
    char '/'
    char '\\'
    sp
    lab2 <- label
    char '.'
    sp
    return $ IntersRule $ Intersection lab lab1 lab2

-- | 'FullRule's start with @'|'@. Example:
--
-- > | L : f(X1, X2) , @a : X1 => @b : foo .
fullRule :: Parser GeneralRule
fullRule =
    do
    cls <- sepBy flatClause (char ',' >> sp)
    char '='
    char '>'
    sp
    goal <- fullClause
    char '.'
    sp
    return $ FullRule $ cls :=> goal

-- | Either a 'intersRule' or a 'fullRule'.
barRule :: Parser GeneralRule
barRule =
    do
    char '|'
    sp
    intersRule <|> fullRule

-- | 'SimpleRule's do /not/ start with @'|'@. Example:
--
-- >  f(X1, X2) => X2 .
simpleRule :: Parser GeneralRule
simpleRule =
    do
    t1 <- simpleTerm
    char '='
    char '>'
    sp
    t2 <- simpleTerm
    char '.'
    sp
    return $ SimpleRule $ [Clause simpleLoc t1] :=> (Clause simpleLoc t2)

rule :: Parser GeneralRule
rule = barRule <|> simpleRule

-- | Grammar production parser. Example:
-- 
-- > @a : f(@b,@c) , 2 , @b .
production :: Parser (Label, [Term])
production =
    do
    l <- label
    char ':'
    sp
    ts <- sepBy groundTerm (char ',' >> sp)
    char '.'
    sp
    return (l,ts)

-- | Parser for the whole @.rules@ file.
rulesFile :: Parser ([Rule],[Rule],[Intersection],[(Label,[Term])])
rulesFile =
    do
    sp
    rules <- many rule
    char '%'
    char '%'
    sp
    grammar <- many production
    let (simple,full,inters) = splitGeneralRules rules
    return (simple, full, inters, grammar)

splitGeneralRules :: [GeneralRule] -> ([Rule],[Rule],[Intersection])
splitGeneralRules grs =
    let aux (SimpleRule r : rs) (srs,frs,irs) = aux rs (r:srs,frs,irs)
	aux (FullRule   r : rs) (srs,frs,irs) = aux rs (srs,r:frs,irs)
	aux (IntersRule r : rs) (srs,frs,irs) = aux rs (srs,frs,r:irs)
	aux []                  (srs,frs,irs) = (srs,frs,irs)
	(sr,fr,ir) = aux grs ([],[],[])
	in (reverse sr, reverse fr, reverse ir)
