{-# OPTIONS_GHC -cpp #-}
-- #ignore-exports
module Renamer
    ( Head, Variable, Label(..), Loc(..), Term(..)
    , Clause(..), Rule(..), Intersection(..), Renamer
    , Subst(..)
    , idSubst, mergeSubst, mergeSubsts, applySubst, applySubstTerm
    , intersectionLabel, intersectionHead, resultLabel, resultHead
    , reservedLabels
    , isPlain
    , succHead
    , minTerm
    , emptyRenamer
    , rename
    , occursInRules, occursInIntersections
    , realLabel, realHead, realVariable
    , renameLabel, renameLabel', renameTerm
    , escapeLabel, escapeTerm
    , pprTerm, pprSaveTerm
    ) where

import qualified Data.Map as Map
import Data.List (nub, sort, (\\))
import qualified Syntax as S

newtype Label = Label { fromLabel :: Int } deriving (Eq,Ord,Show)
newtype Head = Head { fromHead :: Int } deriving (Eq, Ord, Show)
newtype Variable = Variable Int deriving (Eq, Ord, Show)

intersectionLabel :: Label
intersectionLabel = Label 0

resultLabel :: Label
resultLabel = Label 1

reservedLabels :: Int
reservedLabels = 2

intersectionHead :: Head
intersectionHead = Head 2

resultHead :: Head
resultHead = Head 3

reservedIDs :: Int
reservedIDs = 4 

type Renamer = Map.Map Int String

emptyRenamer :: Renamer
emptyRenamer = Map.empty

data Term 
    = Var ! Variable
    | Lab ! Label
    | App ! Head ! [ Term ]
      deriving (Show, Ord, Eq)

isPlain :: Term -> Bool
isPlain (Var _) = False
isPlain (Lab _) = True
isPlain (App _ ts) = let isLabel (Lab _) = True
			 isLabel _       = False
			 in all isLabel ts

succHead :: Head -> Head
succHead (Head n) = Head $ succ n

minTerm :: Term
minTerm = Lab $ Label 0

-- | 'Loc'ations.
data Loc
    = LVar ! Variable
    | LLab ! Label 
      deriving (Eq, Show, Ord)

-- | A 'Clause' is meant to be a pattern for matching 'Term's beloging to
--   a 'Grammar.Grammar'.
data Clause 
    = Clause ! Loc ! Term

-- | The 'Rule's define the 'Term' rewriting system. The implied meaning is
--   \"whenever the left hand side has a match in the 'Grammar.Grammar', 
--   add the right hand side to the 'Grammar.Grammar'\".
data Rule 
    = ! [Clause] :=> ! Clause

-- | A special 'IntersectionRule', defining a 'Label' set as the intersection
--   of two sets.  This could be modeled by a set of 'Rule's, once for each
--   term head.  See also 'Grammar.intersect'.
data Intersection
    = Intersection ! Label ! Label ! Label

-- | Utility.
converseMap :: (Ord e,Ord k) => Map.Map k e -> Map.Map e k
converseMap = Map.fromList . map (\(x,y) -> (y,x)) . Map.assocs 

-- | The main routine.
rename :: [(S.Label,[S.Term])]
       -> [S.Rule]
       -> [S.Intersection]
       -> (Renamer            -- subst
	  ,[(Label, [Term])]  -- new productions
	  ,[Rule]             -- new rules
	  ,[Intersection]     -- new inters
	  ,Int                -- next free label
	  )
rename prods rules inters =
    let namesInTerm S.MinTerm      = []
        namesInTerm (S.Var v)    = [S.fromVariable v]
        namesInTerm (S.Lab l)    = [S.fromLabel l]
	namesInTerm (S.App h ts) = S.fromHead h : concatMap namesInTerm ts
        namesInLoc (S.LLab l) = [S.fromLabel l]
	namesInLoc (S.LVar v) = [S.fromVariable v]
        namesInClause (S.Clause loc t) = namesInLoc loc ++ namesInTerm t
	namesInRule (cs S.:=> c) = concatMap namesInClause (c:cs)
	namesInRules = concatMap namesInRule rules
	namesInInters (S.Intersection l1 l2 l3) = map S.fromLabel [l1,l2,l3]
	namesInInterss = concatMap namesInInters inters
        namesInProd (l,ts) = S.fromLabel l : concatMap namesInTerm ts
        namesInProds = concatMap namesInProd prods
	names' = nub $ namesInProds ++ namesInRules ++ namesInInterss
	-- We sort to get consistent ordering.
	names = sort $ names' \\ [S.fromLabel S.intersectionLabel,
                                  S.fromHead S.intersectionHead,
				  S.fromLabel S.resultLabel,
				  S.fromHead S.resultHead]
	assoc :: [(String,Int)]
	assoc = 
            (S.fromLabel S.intersectionLabel, fromLabel intersectionLabel) :
            (S.fromHead S.intersectionHead  , fromHead intersectionHead) :
	    (S.fromLabel S.resultLabel      , fromLabel resultLabel) :
	    (S.fromHead S.resultHead        , fromHead resultHead) :
	    zip names [reservedIDs..]
	dir = Map.fromList $ assoc
	inv = converseMap dir
        d :: String -> Int
        d s = case Map.lookup s dir of
              Just x -> x
              Nothing -> error $ "rename: " ++ s ++ " not in the map."
	subst :: S.Term -> Term
	subst S.MinTerm = error "rename: minterm"
	subst (S.Var v) = Var $ Variable $ d $ S.fromVariable v
	subst (S.Lab l) = Lab $ Label $ d $ S.fromLabel l
	subst (S.App h ts) = App (Head $ d $ S.fromHead h) $ 
                                 map subst ts
	substProd :: (S.Label,[S.Term]) -> (Label,[Term])
	substProd (l,ts) = (Label $ d (S.fromLabel l) , map subst ts)
	substLoc :: S.Loc -> Loc
	substLoc (S.LVar v) = LVar $ Variable $ d $ S.fromVariable v
	substLoc (S.LLab l) = LLab $ Label $ d $ S.fromLabel l
	substClause :: S.Clause -> Clause
	substClause (S.Clause loc term) = Clause (substLoc loc) (subst term)
	substRule :: S.Rule -> Rule
	substRule (cs S.:=> c) = (map substClause cs) :=> substClause c
	substInters :: S.Intersection -> Intersection
	substInters (S.Intersection l1 l2 l3) = 
	    Intersection 
            (Label $ d (S.fromLabel l1)) 
            (Label $ d (S.fromLabel l2))
            (Label $ d (S.fromLabel l3))
	in (inv,
	   map substProd prods, 
	   map substRule rules, 
	   map substInters inters,
	   length names + reservedIDs)

-- | Ugly hack to inject Int ordering into a String.
prepend0 :: String -> String
prepend0 s = case length s of
	     0 -> "0000"
	     1 -> "000"++s
	     2 -> "00"++s
	     3 -> "0"++s
	     4 -> s
	     _ -> error "prepend0: too big"

-- | The true user-readable name of a 'Label'.
realLabel :: Renamer -> Label -> String
realLabel r (Label n) = case Map.lookup n r of
	                Just lab -> lab
	                Nothing -> '$' : prepend0 (show n)

-- | The true user-readable name of a 'Head'.
realHead :: Renamer -> Head -> String
realHead r (Head n) = case Map.lookup n r of 
	              Just h -> h
	              Nothing -> error "realHead"

-- | The true user-readable name of a 'Variable'.
realVariable :: Renamer -> Variable -> String
realVariable r (Variable n) = case Map.lookup n r of
	                      Just lab -> lab
	                      Nothing -> error "realVariable"

-- | Slow renaming of 'Syntax.Label'.
renameLabel :: Renamer -> S.Label -> Maybe Label
-- Special case: @$nnnn labels are converted to $nnnn. This is to allow
-- the user to enter them in interactive mode, since the Parser does not
-- handle them. (Still hackish, though.)
renameLabel _ (S.Label ('@':'$':s)) =
    case reads s of
      [(n,"")] -> Just $ Label n
      _        -> Nothing
renameLabel r (S.Label s) = fmap Label $ Map.lookup s $ converseMap r

-- | Variant of 'renameLabel'.
renameLabel' :: Renamer -> String -> Maybe Label
renameLabel' r s = renameLabel r (S.Label s)

-- | Slow renaming of 'Syntax.Head'.
renameHead :: Renamer -> S.Head -> Maybe Head
renameHead r (S.Head s) = fmap Head $ Map.lookup s $ converseMap r

-- | Slow renaming of 'Syntax.Variable'.
renameVariable :: Renamer -> S.Variable -> Maybe Variable
renameVariable r (S.Variable s) = fmap Variable $ Map.lookup s $ converseMap r

-- | Slow renaming of 'Syntax.Term'.
renameTerm :: Renamer -> S.Term -> Maybe Term
renameTerm _ S.MinTerm = error "renameTerm: MinTerm"
renameTerm r (S.Lab l) = fmap Lab $ renameLabel r l
renameTerm r (S.Var v) = fmap Var $ renameVariable r v
renameTerm r (S.App h ts) = do h'  <- renameHead r h
                               ts' <- mapM (renameTerm r) ts
                               return $ App h' ts'

-- * Substitutions

-- TODO Check if FiniteMap is better, faster, etc.
newtype Subst = Subst [(Variable, Term)] 
                deriving Show

-- | The identity substitution.
idSubst :: Subst
idSubst = Subst []

-- | Glues two disjoint substitutions together.
mergeSubst :: Subst -> Subst -> Subst
Subst x `mergeSubst` Subst y = Subst $ x ++ y

-- | Glues many disjoint substitutions together.
mergeSubsts :: [Subst] -> Subst
mergeSubsts s = Subst $ concat [ l | Subst l <- s ]

applySubstTerm :: Subst -> Term -> Term
(Subst u) `applySubstTerm` tv@(Var v) =
    case lookup v u of
        Just t -> t
	Nothing -> tv
_ `applySubstTerm` t@(Lab _) = t
s `applySubstTerm` (App h ts) = App h (map (applySubstTerm s) ts)

-- | Apply 'Subst' to a 'Clause'. The 'Loc'ation in the result must be a
--   'Label'.
applySubst :: Subst -> Clause -> Clause
applySubst s@(Subst u) (Clause l t) =
    let l' :: Loc
	l' = case l of
		 LLab _ -> l
		 LVar v -> 
		     case lookup v u of
		         Nothing -> l
		         Just (Lab lab) -> LLab lab
			 _ -> error ("applySubst: subst must replace " ++
				     "location with location. " ++
				     show s ++ " var " ++ show v)
	in Clause l' (s `applySubstTerm` t)

-- * Occurrence
occursInRules :: Label -> [Rule] -> Bool
l `occursInRules` rs = any (l `occursInRule`) rs

occursInRule :: Label -> Rule -> Bool
l `occursInRule` (cs :=> c) = any (l `occursInClause`) (c:cs)

occursInClause :: Label -> Clause -> Bool
l `occursInClause` (Clause loc term) = l `occursInLoc` loc || 
                                       l `occursInTerm` term

occursInLoc :: Label -> Loc -> Bool
l `occursInLoc` lab = LLab l == lab

occursInTerm :: Label -> Term -> Bool
l `occursInTerm` (Lab l')   = l == l'
l `occursInTerm` (App _ ts) = any (l `occursInTerm`) ts
_ `occursInTerm` _          = False

occursInIntersections :: Label -> [Intersection] -> Bool
l `occursInIntersections` is = any (l `occursInIntersection`) is

occursInIntersection :: Label -> Intersection -> Bool
l `occursInIntersection` (Intersection l1 l2 l3) = l==l1 || l==l2 || l==l3

-- * Pretty Printing

-- | Pretty print a 'Term'.
pprTerm :: Renamer -> Term -> String
pprTerm r (Var v) = realVariable r v
pprTerm r (Lab l) = realLabel r l
pprTerm r (App h []) = realHead r h
pprTerm r (App h ts) = realHead r h ++ "(" ++ pprTerms r ts ++ ")"

pprTerms :: Renamer -> [Term] -> String
pprTerms _ [] = ""
pprTerms r [t] = pprTerm r t
pprTerms r (t:ts) = pprTerm r t ++ ", " ++ pprTerms r ts

-- | Pretty print a term in a format suitable for saving in a @.rules@ file.
pprSaveTerm :: Renamer -> Term -> String
pprSaveTerm r (Var v) = realVariable r v
pprSaveTerm r (Lab l) = escapeName $ realLabel r l
pprSaveTerm r (App h []) = realHead r h
pprSaveTerm r (App h ts) = realHead r h ++ "(" ++ pprSaveTerms r ts ++ ")"

pprSaveTerms :: Renamer -> [Term] -> String
pprSaveTerms _ [] = ""
pprSaveTerms r [t] = pprSaveTerm r t
pprSaveTerms r (t:ts) = pprSaveTerm r t ++ ", " ++ pprSaveTerms r ts

-- * Escaping

-- | Prepend @\'\@\'@ to names staring with @\'\$\'@ or @\'%\'@, in order 
--   to transform names for internal use only into user available user names.
--
--   [NOTE] The computed closure /is/ affected by label renaming: see
--   'Main.addPlainTermSomewhere' for details. 'escapeName' should try to
--   rename \"better\" labels into smaller (@<@) ones.
--
--   We try to keep @\'%\'@ names small, by prefixing them with @\@a_@.
--   Similarly, we prefix @\'\$\'@ names with @\@z_@.
escapeName :: String -> String
escapeName l@('$':_) = "@z_" ++ l
escapeName l@('%':_) = "@a_" ++ l
escapeName l | S.Label l == S.intersectionLabel = l
escapeName l | S.Label l == S.resultLabel = l
escapeName l@('@':_) = l
escapeName _         = error "escapeName: unknown name"

escapeLabel :: Renamer -> Label -> S.Label
escapeLabel r l = S.Label $ escapeName $ realLabel r l

escapeTerm :: Renamer -> Term -> S.Term
escapeTerm r (Lab l) = S.Lab $ escapeLabel r l
escapeTerm r (App h ts) = S.App (S.Head $ realHead r h) $ map (escapeTerm r) ts
escapeTerm r (Var v) = S.Var $ S.Variable $ realVariable r v
