{-# OPTIONS_GHC -cpp #-}
module Syntax where

-- | A string used for a variable name. 
--   Names starting with @'!'@ are reserved for internal use.
newtype Variable 
    = Variable { fromVariable :: String } deriving (Eq, Show, Ord)

-- | A string used for a term head name.
newtype Head 
    = Head { fromHead :: String } deriving (Eq, Show, Ord)

-- | A string used for a label name (includes the @\'\@\'@).
--   Names starting with @\"\@!\"@ are reserved for internal use.
newtype Label 
    = Label { fromLabel :: String } deriving (Eq, Show, Ord)

-- | 'Term's. We often use the following definitions:
--
-- [depth] The maximum number of 'App' nestings in a 'Term'. Note that
--         @foo@, @foo(X)@, and @foo(\@lab)@ have depth one, while @foo(bar)@
--         has depth two, and @X@ and @\@lab@ have depth zero.
--
-- [ground term] A term without variables.
--
-- [simple term] A term without labels.
--
-- [pure term]   A /simple/ /ground/ term.
--
-- [plain term]  A /ground/ term of depth <= 1.
data Term 
    = MinTerm                    -- ^ Used to easily define orderings.
    | Var ! Variable
    | Lab ! Label
    | App ! Head ! [ Term ]
      deriving (Eq, Show, Ord)

-- | 'Loc'ations.
data Loc
    = LVar ! Variable 
    | LLab ! Label 
      deriving (Eq, Show, Ord)

-- | A 'Clause' is meant to be a pattern for matching 'Term's beloging to
--   a 'Grammar.Grammar'.
data Clause 
    = Clause ! Loc ! Term

-- | The 'Rule's define the 'Term' rewriting system. The implied meaning is
--   \"whenever the left hand side has a match in the 'Grammar.Grammar', 
--   add the right hand side to the 'Grammar.Grammar'\".
data Rule 
    = ! [Clause] :=> ! Clause

-- | A special 'IntersectionRule', defining a 'Label' set as the intersection
--   of two sets.  This could be modeled by a set of 'Rule's, once for each
--   term head.  See also 'Grammar.intersect'.
data Intersection
    = Intersection ! Label ! Label ! Label

-- * Special Constants

-- | A 'Label' whose set over-approximates the intersection between sets.
--   This is used internally: see also 'Grammar.computeIntersection'.
intersectionLabel :: Label
intersectionLabel = Label "@!Inters"

-- | A 'Head' that is used in the set of 'intersectionLabel' in order to
--   over-approximate the intersection between sets.
--   This is used internally: see also 'Grammar.computeIntersection'.
intersectionHead :: Head
intersectionHead = Head "cons"

-- | A 'Label' where is stored the result fo the analisys, by convention.
resultLabel :: Label
resultLabel = Label "@zzResult"

-- | A flag that, if present in 'resultLabel', tells the analysis failed.
resultHead :: Head
resultHead = Head "!FAIL!"

-- | The 'Loc'ation implied in a 'Parser.SimpleRule'.
simpleLoc :: Loc
simpleLoc = LVar $ Variable "!L"

-- * Substitutions

-- TODO Check if FiniteMap is better, faster, etc.
newtype Subst = Subst [(Variable, Term)] 
                deriving Show

applySubstTerm :: Subst -> Term -> Term
(Subst u) `applySubstTerm` tv@(Var v) =
    case lookup v u of
        Just t -> t
	Nothing -> tv
_ `applySubstTerm` t@(Lab _) = t
s `applySubstTerm` (App h ts) = App h (map (applySubstTerm s) ts)
_ `applySubstTerm` MinTerm = MinTerm

-- | Apply 'Subst' to a 'Clause'. The 'Loc'ation in the result must be a
--   'Label'.
applySubst :: Subst -> Clause -> Clause
applySubst s@(Subst u) (Clause l t) =
    let l' :: Loc
	l' = case l of
		 LLab _ -> l
		 LVar v -> 
		     case lookup v u of
		         Nothing -> l
		         Just (Lab lab) -> LLab lab
			 _ -> error ("applySubst: subst must replace " ++
				     "location with location. " ++
				     show s ++ " var " ++ show v)
	in Clause l' (s `applySubstTerm` t)

-- * Utilities
isPlain :: Term -> Bool
isPlain MinTerm    = True
isPlain (Lab _)    = True
isPlain (App _ ts) = let isLabel :: Term -> Bool
			 isLabel (Lab _) = True
			 isLabel _       = False
			 in all isLabel ts
isPlain _          = False

succHead :: Head -> Head
succHead (Head n) = Head $ n ++ "\NUL"

-- * Pretty Printing

-- | Pretty print a 'Term'.
pprTerm :: Term -> String
pprTerm (Var (Variable v)) = v
pprTerm (Lab (Label l)) = l
pprTerm (App (Head h) []) = h
pprTerm (App (Head h) ts) = h ++ "(" ++ pprTerms ts ++ ")"
pprTerm MinTerm = "<<MinTerm>>"

pprTerms :: [Term] -> String
pprTerms [] = ""
pprTerms [t] = pprTerm t
pprTerms (t:ts) = pprTerm t ++ ", " ++ pprTerms ts

-- | Pretty print a 'Clause'.
pprClause :: Clause -> String
pprClause (Clause (LLab (Label l)) term)    = l ++ " : " ++ pprTerm term
pprClause (Clause (LVar (Variable v)) term) = v ++ " : " ++ pprTerm term

pprClauses :: [Clause] -> String
pprClauses [] = ""
pprClauses [c] = pprClause c
pprClauses (c:cs) = pprClause c ++ ", " ++ pprClauses cs

-- | Pretty print a 'Rule'.
pprRule :: Rule -> String
pprRule (cs :=> goal) = pprClauses cs ++ " => " ++ pprClause goal

-- | Pretty print 'Parser.SimpleRule's and 'Parser.FullRule's , using 
--   user-friendly syntax.
pprGeneralRule :: Rule -> String
pprGeneralRule ([Clause loc1 term1] :=> (Clause loc2 term2)) 
	       | loc1 == simpleLoc && loc2 == simpleLoc  =
    pprTerm term1 ++ " => " ++ pprTerm term2
pprGeneralRule otherRule = '|':' ': pprRule otherRule
