{-# OPTIONS -cpp #-}
-- #ignore-exports
module TypeCheck 
    ( typeCheckRulesAndGrammar
    , typeErrorMsg
    )
where

import qualified Data.Map as Map
import Data.Maybe
import Control.Monad.State

import Syntax

-- * Types

-- | The type of a variable (with \'absint\' ordering).
data Type 
    = TypeLabel  -- ^ A variable that will be bound to labels, only.
    | TypeTerm   -- ^ A variable that can be bound to any term.
    deriving (Eq,Ord)

-- | While pattern matching, the definition state of a variable.
--   Used to spot doubly defined variables ( malformed patterns as @f(X,X)@ ).
data Def 
    = Defined    -- ^ Variable was defined by an earlier match.
    | Defining   -- ^ Variable was defined by the current match.

-- | The state holding all the information about variables and term heads.
data TypeEnv = 
    TypeEnv 
    { tyDef :: ! (Map.Map Variable (Type, Def))  
                                       -- ^ Type and definition of variables.
    , arity :: ! (Map.Map Head Int)  -- ^ Arity of term heads.
    }

-- * State of the Typechecker

-- | The actual state monad. Represents a computation having a 'TypeEnv' as a 
--   state, but that can fail with a 'TypeError'.
type TypeSt = StateT TypeEnv TypeCheckResult

-- | Forget all about variables (but remember term heads information)
forgetVariables :: TypeSt ()
forgetVariables = modify (\st -> st{ tyDef = Map.empty })

-- | Make 'Defining' variables become 'Defined'.
fixVariables :: TypeSt ()
fixVariables = 
    let change :: (Type, Def) -> (Type, Def)
	change (t, Defining) = (t, Defined)
	change other         = other
	in modify (\st -> st{ tyDef = Map.map change (tyDef st) })

-- | Check a term 'Head'.
--
--  * If already known, but the arity disagrees, fail.
--
--  * Otherwise, remember the arity.
checkHead :: Head -> Int -> TypeSt ()
checkHead h a =
    do
    st <- get
    case h `Map.lookup` arity st of
        Nothing           -> put $ st{ arity = Map.insert h a (arity st) }
        Just a' | a == a' -> return ()
        Just a'           -> typeError $ MultipleArity h a' a

-- | Check a 'Variable' occurring in a pattern.
--
--   * If already known, but is not 'Defined' or the type does not match, fail.
--
--   * Otherwise, mark it as 'Defining' and remember the type.
checkVar :: Variable -> Type -> TypeSt ()
checkVar v ty =
    do
    st <- get
    case v `Map.lookup` tyDef st of
        Just (_, Defining) ->  -- Already defined by the same pattern.
            typeError $ MultipleDef v
        Just (ty', Defined) -> -- Already defined by a previous pattern.
            if ty' <= ty
               then return ()
               else typeError $ LabelRequired v
        Nothing -> -- Never seen before: first definition.
            put $ st{ tyDef = Map.insert v (ty, Defining) (tyDef st) }

-- | Define a 'Variable' using 'TypeTerm' and 'Defining'. Do /not/ perform any
--   check.
definingVar :: Variable -> TypeSt ()
definingVar v = modify (\st -> 
    st{ tyDef = Map.insert v (TypeTerm, Defining) (tyDef st) })

-- | If a variable is not 'Defined', fail.
useVar :: Variable -> TypeSt ()
useVar v =
    do
    tD <- gets tyDef
    case v `Map.lookup` tD of
        Just (_,Defined) -> return ()
        _                -> typeError $ NoDef v

-- * Checking the Goal ('Clause' head)

-- | Check the goal 'Term'.
checkGoalTerm :: Term -> TypeSt ()
checkGoalTerm MinTerm = error "checkGoalTerm: MinTerm"
checkGoalTerm (Lab _) = return ()
checkGoalTerm (Var v) = useVar v
checkGoalTerm (App h ts) = 
    do
    checkHead h $ length ts
    mapM_ checkGoalTerm ts

-- | Check a goal.
checkGoal :: Clause -> TypeSt ()
checkGoal (Clause loc term) =
    do
    case loc of
        LVar v -> do useVar v ; checkVar v TypeLabel
        LLab _ -> return ()
    checkGoalTerm term

-- * Checking 'FullRule's

-- | Check the toplevel pattern of a 'FullRule'.
checkToplevelPattern :: Term -> TypeSt ()
checkToplevelPattern MinTerm = error "checkToplevelPattern: MinTerm"
checkToplevelPattern (Lab _) = return ()
checkToplevelPattern (Var v) = checkVar v TypeTerm
checkToplevelPattern term@(App h ts) =
    do
    checkHead h $ length ts
    mapM_ (checkInnerPattern term) ts

-- | Check a pattern that was inside a @f(-)@ context in a 'FullRule'.
checkInnerPattern :: Term -> Term -> TypeSt ()
checkInnerPattern _ MinTerm = error "checkInnerPattern: MinTerm"
checkInnerPattern _ (Lab _) = return ()
checkInnerPattern _ (Var v) = checkVar v TypeLabel
checkInnerPattern term (App _ _) = 
    -- Pattern is of the form @f(..., g(-), ...)@.
    typeError $ TermTooDeep term

-- | Check a 'FullRule' 'Clause'.
checkClause :: Clause -> TypeSt ()
checkClause (Clause loc term) =
    do
    case loc of
        LVar v -> checkVar v TypeLabel
        LLab _ -> return ()
    checkToplevelPattern term
    fixVariables

checkFullRule :: Rule -> TypeSt ()
checkFullRule rule@(cls :=> goal) =
    setRuleOnError rule $
        do
	forgetVariables -- ... but preserve arity information!
	mapM_ checkClause cls
	checkGoal goal

-- * Checking 'SimpleRule's

-- | Check the pattern inside a 'SimpleRule'.
checkSimplePattern :: Term -> TypeSt ()
checkSimplePattern MinTerm = error "checkSimplePattern: MinTerm"
checkSimplePattern (Lab _) = error "checkSimplePattern: Lab"
checkSimplePattern (Var v) = definingVar v
    -- Define the variable, without checking for 'MultipleDef'.  
checkSimplePattern (App h ts) =
    do
    checkHead h $ length ts
    mapM_ checkSimplePattern ts

-- | Check a 'SimpleRule' for arity consistency, and undefined variables.
checkSimpleRule :: Rule -> TypeSt ()
checkSimpleRule rule@([Clause _ term] :=> goal) =
    setRuleOnError rule $
        do
        let internalVar = -- The variable implied by 'SimpleRule's.
                case simpleLoc of 
                    LVar v -> v
                    _      -> error "checkSimpleRule: internal error 1"
        forgetVariables -- ... but preserve arity information!
        checkVar internalVar TypeLabel
	checkSimplePattern term
	fixVariables
        checkGoal goal
checkSimpleRule _ = error "checkSimpleRule: internal error 2"

-- * Main Routines.

-- | Check 'SimpleRule's and 'FullRule's,
typeCheckRules :: [Rule] -> [Rule] -> TypeSt ()
typeCheckRules simpleRules fullRules =
    do
    mapM_ checkSimpleRule simpleRules
    mapM_ checkFullRule fullRules

-- | Check for arity consistency.
typeCheckGrammar :: [(Label,[Term])] -> TypeSt ()
typeCheckGrammar grammar =
    do
    let checkProduction (lab,terms) =
            let checkTerm :: Term -> TypeSt ()
                checkTerm term =
                    -- Hack to tell typeErrorMsg this is a grammar error.
                        setProductionOnError lab term $
                            checkGoalTerm term
                in mapM_ checkTerm terms
    forgetVariables
    mapM_ checkProduction grammar

-- | The main type checking routine. Return 'Nothing' on success.
--   On error, the 'Rule' premiss is empty iff the error was in the grammar.
typeCheckRulesAndGrammar :: [Rule] -> [Rule] -> [(Label,[Term])] -> 
                            Maybe (Rule, TypeError)
typeCheckRulesAndGrammar simpleRules fullRules grammar =
    let initialState :: TypeEnv
        initialState = TypeEnv { tyDef = Map.empty 
                               , arity = Map.singleton intersectionHead 2 
                               }
        checkAll :: TypeSt ()
	checkAll =
	    do
            typeCheckRules simpleRules fullRules
            typeCheckGrammar grammar
	in case evalStateT checkAll initialState of
	       TypeCheck _   -> Nothing
	       TypeError r e -> Just (r,e)    


-- * Type Errors

data TypeError 
    = TermTooDeep Term       -- ^ Maximum term depth is one in 'FullRule's.
    | LabelRequired Variable -- ^ A label or a label-valued variable was 
                             --   expected.
    | MultipleArity Head     
                    Int Int  -- ^ Term head with two distinct arities found.
    | MultipleDef Variable   -- ^ A pattern defined a variable more than once.
    | NoDef Variable         -- ^ A variable occurs only in the goal.

-- | A simple error monad that propagates errors.
data TypeCheckResult a 
    = TypeCheck a                 -- ^ Type check succeded.
    | TypeError Rule TypeError    -- ^ Type check failed for a rule.

instance Monad TypeCheckResult where
    return x = TypeCheck x
    (TypeCheck x  ) >>= f = f x
    (TypeError r e) >>= _ = TypeError r e

-- | User-friendly error messages.
typeErrorMsg :: Rule -> TypeError -> String
typeErrorMsg rule err =
    let msg :: TypeError -> String
        msg (TermTooDeep term) = 
	    "Term " ++ pprTerm term ++ " is too deep."
	msg (LabelRequired (Variable vName)) = 
	    "Variable '" ++ vName ++ "' must be bound to a label."
	msg (MultipleArity (Head hName) old new) = 
	    "Term head '" ++ hName ++ "' was used both with arity " ++ 
	    show old ++ " and " ++ show new ++ "."
	msg (MultipleDef (Variable vName)) = 
	    "Pattern defines variable '" ++ vName ++ "' multiple times."
        msg (NoDef (Variable vName)) =
            "Variable '" ++ vName ++ "' is used but never defined."

        handleRule :: Rule -> String
        handleRule ([] :=> (Clause (LLab (Label ('$':lName))) term)) =
            -- Special case: grammar production.
            "automaton transition\n    " ++ lName ++ " : " ++ pprTerm term
        handleRule r = 
            "rewriting rule\n    " ++ pprGeneralRule r
    in "Type error in " ++ handleRule rule ++ " .\n" ++ msg err

-- * Utilities

-- | Smart constructor.
typeError :: TypeError -> TypeSt ()
typeError te = lift $ TypeError undefined te

{- Utility: (little hack)
   Behave as action, but trap all TypeError's. On TypeError, rethrow
   the same error with another 'Rule' instead.

   This is used to avoid passing the error-related rule to many subroutines, 
   or storing it into the state. Subroutines can simply use the smart 
   constructor 'typeError' (which uses 'undefined' for the 'Rule').
-}	
setRuleOnError :: Rule -> TypeSt a -> TypeSt a
setRuleOnError rule action =
    do
    result <- gets $ runStateT action
    case result of
        TypeError _ e    -> lift $ TypeError rule e
	TypeCheck (x,st) -> do put st ; return x

setProductionOnError :: Label -> Term -> TypeSt a -> TypeSt a
setProductionOnError (Label lName) term =
    -- Prepend '$' to the label name, so we can check for this later. (hack)
    setRuleOnError ([] :=> (Clause (LLab $ Label $ '$' : lName) term))
