-- #ignore-exports
module Main (main) where

import System
import System.IO
import Data.List
import Text.ParserCombinators.Parsec (parse)
import qualified Data.Map as Map

import Syntax
import Parser

-- | The type of the analysis.
data Analysis = Std | InOut | Apoc

-- * Standard (CFA-style) analysis.

-- | Static environment.
type Env = Map.Map Name Label

-- | Empty initial environment.
rho0 :: Env
rho0 = Map.empty

-- | 'Name' lookup in 'Env'.
lookupName :: Name -> Env -> Label
lookupName n@(Name s) r =
    maybe (error $ "undefined " ++ s) id $ Map.lookup n r

-- | Instantiate all the 'Var's in a 'Term'.
applyEnv :: Env -> Term -> Term
applyEnv r (App h ts) = App h (map (applyEnv r) ts)
applyEnv r (Var n)    = Lab $ lookupName n r
applyEnv _ term       = term

-- | Generate the 'Production's for the static analysis of the 'Proc'ess.
genStd :: Position      -- ^ The address of the process
       -> Label         -- ^ The 'Label' representing the input\/output
       -> Env           -- ^ Binding from 'Name's to 'Label's
       -> Proc          -- ^ The (sub)process we are analyzing
       -> [Production]  -- ^ The grammar
genStd _ _  _ Nil = []
genStd t io r (In n cont) =
    let	r' = Map.insert n io r
	in Remark (pprBinding n io) : 
           genStd (next t) io r' cont
genStd t io r (Out term cont) =
    P io (applyEnv r term) : genStd (next t) io r cont
genStd t io r (New n cont) =
    let lab  = newLabelForName t n "_val"
	lab' = newLabelForName t n ""
	r' = Map.insert n lab' r
	in P lab (constOfName n "") : P lab (App "next" [Lab lab]) :
	   P lab' (App "val" [Lab lab]) :
	   genStd (next t) io r' cont
genStd t io r (Par p1 p2) = genStd (left t) io r p1 ++ genStd (right t) io r p2
genStd t io r (Repl cont) = genStd (next t) io r cont
genStd t io r (Let n term cont) =
    let lab = newLabelForName t n ""
        r' = Map.insert n lab r
	in P lab (applyEnv r term) :
	   genStd (next t) io r' cont
genStd t io r (Match n1 n2 cont) =
    let lab  = newLabelForName t n1 ""
	lab1 = lookupName n1 r
	lab2 = lookupName n2 r
        r' = Map.insert n2 lab $ Map.insert n1 lab r
	in Remark (pprBinding n1 lab) : Remark (pprBinding n2 lab) : 
           Inters lab lab1 lab2 :
	   genStd (next t) io r' cont
genStd _t _io _r (Check _) =
    [ Remark "WARNING : ignoring chk in standard mode !"  ]

-- * /input\/output/ analysis

-- | Generate the 'Production's for the static analysis of the 'Proc'ess.
genInOut :: Position      -- ^ The address of the process
         -> Label         -- ^ The 'Label' representing the input
         -> Label         -- ^ The 'Label' representing the output
         -> Env           -- ^ Binding from 'Name's to 'Label's
         -> Proc          -- ^ The (sub)process we are analyzing
         -> [Production]  -- ^ The grammar
genInOut _ _   _   _ Nil = []
genInOut t inp out r (In n cont) =
    let	r' = Map.insert n inp r
	in Remark (pprBinding n inp) : 
           genInOut (next t) inp out r' cont
genInOut t inp out r (Out term cont) =
    P out (applyEnv r term) : genInOut (next t) inp out r cont
genInOut t inp out r (New n cont) =
    let lab  = newLabelForName t n "_val"
	lab' = newLabelForName t n ""
	r' = Map.insert n lab' r
	in P lab (constOfName n "") : P lab (App "next" [Lab lab]) :
	   P lab' (App "val" [Lab lab]) :
	   genInOut (next t) inp out r' cont
genInOut t inp out r (Par p1 p2) =
    let (t1, t2) = (left t, right t)
        inp1 = newLabel t1 "input"
	inp2 = newLabel t2 "input"
        out1 = newLabel t1 "output"
	out2 = newLabel t2 "output"
	in P inp1 (Lab inp)  : P inp1 (Lab out2) :
	   P inp2 (Lab inp)  : P inp2 (Lab out1) :
	   P out  (Lab out1) : P out  (Lab out2) :
	   (genInOut (left t)  inp1 out1 r p1 ++
	    genInOut (right t) inp2 out2 r p2)
genInOut t inp out r (Repl cont) =
    let t1 = next t
        inp1 = newLabel t1 "input"
	out1 = newLabel t1 "output"
	in P inp1 (Lab inp)  : P inp1 (Lab out1) : 
           P out  (Lab out1) :
	   genInOut t1 inp1 out1 r cont
genInOut t inp out r (Let n term cont) =
    let lab = newLabelForName t n ""
        r' = Map.insert n lab r
	in P lab (applyEnv r term) :
	   genInOut (next t) inp out r' cont
genInOut t inp out r (Match n1 n2 cont) =
    let lab  = newLabelForName t n1 ""
	lab1 = lookupName n1 r
	lab2 = lookupName n2 r
        r' = Map.insert n2 lab $ Map.insert n1 lab r
	in Remark (pprBinding n1 lab) : Remark (pprBinding n2 lab) : 
           Inters lab lab1 lab2 :
	   genInOut (next t) inp out r' cont
genInOut _t _inp _out _r (Check _) =
    [ Remark "WARNING : ignoring chk in input/output mode !"  ]

-- * /post-apocalyptic/ analysis.

-- | Generate the 'Production's for the static analysis of the 'Proc'ess.
genApoc :: Position      -- ^ The address of the process
        -> Label         -- ^ The 'Label' representing the input
        -> Label         -- ^ The 'Label' representing the output
        -> Env           -- ^ Binding from 'Name's to 'Label's
        -> Proc          -- ^ The (sub)process we are analyzing
        -> [Production]  -- ^ The grammar
genApoc _ _   _   _ Nil = []
genApoc t inp out r (In n cont) =
    let	r' = Map.insert n inp r
	in Remark (pprBinding n inp) : 
           genApoc (next t) inp out r' cont
genApoc t inp out r (Out term cont) =
    P out (applyEnv r term) : genApoc (next t) inp out r cont
genApoc t inp out r (New n cont) =
    let lab  = newLabelForName t n "_val"
	lab' = newLabelForName t n ""
	r' = Map.insert n lab' r
	in P lab (constOfName n "") : P lab (App "next" [Lab lab]) :
	   P lab' (App "val" [Lab lab]) :
	   genApoc (next t) inp out r' cont
genApoc t inp out r (Par p1 p2) =
    let (t1, t2) = (left t, right t)
        inp1 = newLabel t1 "input"
	inp2 = newLabel t2 "input"
        out1 = newLabel t1 "output"
	out2 = newLabel t2 "output"
	in P inp1 (Lab inp)  : P inp1 (Lab out2) :
	   P inp2 (Lab inp)  : P inp2 (Lab out1) :
	   P out  (Lab out1) : P out  (Lab out2) :
	   (genApoc (left t)  inp1 out1 r p1 ++
	    genApoc (right t) inp2 out2 r p2)
genApoc t inp out r (Repl cont) =
    let t1 = next t
        inp1 = newLabel t1 "input"
	out1 = newLabel t1 "output"
	in P inp1 (Lab inp)  : P inp1 (Lab out1) : 
           P out  (Lab out1) :
	   genApoc t1 inp1 out1 r cont
genApoc t inp out r (Let n term cont) =
    let lab = newLabelForName t n ""
        r' = Map.insert n lab r
	in P lab (applyEnv r term) :
	   genApoc (next t) inp out r' cont
genApoc t inp out r (Match n1 n2 cont) =
    let lab  = newLabelForName t n1 ""
	lab1 = lookupName n1 r
	lab2 = lookupName n2 r
        r' = Map.insert n2 lab $ Map.insert n1 lab r
	in Remark (pprBinding n1 lab) : Remark (pprBinding n2 lab) : 
           Inters lab lab1 lab2 :
	   genApoc (next t) inp out r' cont
genApoc t _inp _out r (Check cont) =
    -- reset in and out to fresh labels
    let inp = newLabel top' "input"
        out = newLabel top' "output"
        in genApoc (next t) inp out r cont

-- * Main

usageMessage :: IO a
usageMessage =
  do
  { putStr $ unlines $
    [ "Usage: generate --<mode> file.proc > file.rules"
    , ""
    , "   where <mode> is one of"
    , "     std    control flow analysis, single state for inputs and outputs"
    , "     inout  as std, with separate input/output states"
    , "     apoc   as inout, with support for chk"
    ]
  ; exitFailure
  }

main :: IO ()
main = do
       args <- getArgs
       (which, filename) <- case args of
	       ["--std"    , f] -> return (Std    , f)
               ["--inout"  , f] -> return (InOut  , f)
               ["--apoc"   , f] -> return (Apoc   , f)
	       _                -> usageMessage
       contents <- readFile filename
       case parse file filename contents of
           Left parseError -> do hPrint stderr parseError ; exitFailure
	   Right (premiss, proc) ->
	       do
               putStrLn $ "# This file was automatically generated from " ++ 
                          filename ++ " ."
               putStrLn $ "# Mode: " ++ head args
	       putStr premiss
	       let comment = "# " ++ pprProc proc ++ "\n"
               let inp = newLabel top "input"
               let out = newLabel top "output"
	       let prods :: [ Production ]
                   prods = case which of
		       Std     -> genStd top (newLabel top "inout") rho0 proc
                       InOut   -> genInOut top inp out rho0 proc
                       Apoc    -> genApoc top inp out rho0 proc
	       putStr $ pprProductions prods comment

