{-# OPTIONS -Wall #-}

-- #ignore-exports
module Parser (file) where

import Prelude hiding (head)

import Control.Monad
import Text.ParserCombinators.Parsec hiding (label)
import Text.ParserCombinators.Parsec.Char

import Syntax

-- | One of some non alphanumeric symbols.
symbol :: Parser Char
symbol = oneOf "\'\";+-*/=^&"

-- | @#-till-newline@ comment parser.
comment :: Parser ()
comment =
    do
    char '#' -- begin of comment
    many $ noneOf ['\n']
    char '\n'
    return ()

-- | Whitespace parser (comments are whitespace).
sp :: Parser ()
sp =
    do
    many ((space >> return ()) <|> comment)
    return ()

-- | Heads start with lowercase char.
head :: Parser String
head =
    do
    u <- lower <|> digit <|> symbol <|> char '!'
    us <- many $ upper <|> lower <|> digit <|> symbol <|> oneOf "$!"
    sp
    return $ u:us

-- | 'Name's start with uppercase char.
name :: Parser Name
name = do
       u <- upper
       us <- many $ upper <|> lower <|> digit <|> symbol <|> oneOf "$!"
       sp
       return $ Name $ u:us

-- | Term parser builder.
termFactory :: Parser Term -> Parser Term
termFactory subTerm = do
       h <- head
       (do
	ts <- between (char '(' >> sp) (char ')' >> sp) $
	          sepBy subTerm (char ',' >> sp)
	return $ App h ts
        ) <|>
        (return $ App h [])

term :: Parser Term
term = (liftM Var name) <|> termFactory term

-- * Processes

input :: Parser Proc
input = do
	mapM_ char "in" >> sp
	n <- name
	char '.'
	sp
	cont <- process
	return $ In n cont

output :: Parser Proc
output = do 
	 mapM_ char "out" >> sp
	 t <- term
	 char '.' >> sp
	 cont <- process
	 return $ Out t cont

new :: Parser Proc
new = do
      mapM_ char "new" >> sp
      n <- name
      char '.' >> sp
      cont <- process
      return $ New n cont

-- | Parallel (includes @nil = ()@).
par :: Parser Proc
par = do
      ps <- between (char '(' >> sp) (char ')' >> sp) $
	      sepBy process (char '|' >> sp) 
      return $ if null ps
	       then Nil
	       else foldr1 Par ps

repl :: Parser Proc
repl = do
       char '!' >> sp
       char '.' >> sp
       p <- process
       return $ Repl p

-- | Let
decl :: Parser Proc
decl = do
       mapM_ char "let" >> sp
       n <- name
       char '=' >> sp
       t <- term
       char '.' >> sp
       cont <- process
       return $ Let n t cont

match :: Parser Proc
match = do
	char '[' >> sp
	n1 <- name
	char '=' >> sp
	n2 <- name
	char ']' >> sp
	char '.' >> sp
	cont <- process
	return $ Match n1 n2 cont

check :: Parser Proc
check = do
	mapM_ char "chk" >> sp
	char '.' >> sp
	cont <- process
	return $ Check cont

process :: Parser Proc
process = 
    new <|> input <|> output <|>
	par <|> repl <|> decl <|> match <|> check

premiss :: Parser String
premiss =
    (char '%' >> 
     ((do char '%' ; sp ; return "") <|> 
      (do s <- premiss ; return $ '%':s)))
    <|> (do c <- anyChar ; s <- premiss ; return $ c:s)

-- | The whole process file. Example:
--
-- @
-- rewriting rule 1 .
-- rewriting rule 1 .
-- rewriting rule 2 .
-- %%
-- process
-- @ 
file :: Parser (String, Proc)
file = do
       pre  <- premiss
       proc <- process
       return (pre,proc)
