-- #ignore-exports
module Syntax where

import qualified Data.List

-- | 'Name's in the process algebra.
newtype Name = Name String      deriving (Show, Eq, Ord)

-- | 'Term' algebra.
data Term = App String [Term]  -- ^ Application
	  | Var Name           -- ^ Variables (only in processes)
	  | Lab Label          -- ^ Labels (only in productions)
	    deriving Show

freeVars :: Term -> [Name]
freeVars (App _ terms) = Data.List.nub $ concatMap freeVars terms
freeVars (Var n)       = [n]
freeVars _             = []

-- | 'Proc'esses of the algebra.
data Proc
    = Nil                    -- ^ @()@
    | In Name Proc           -- ^ @in X . Q@
    | Out Term Proc          -- ^ @out T . Q@
    | New Name Proc          -- ^ @new X . Q@
    | Par Proc Proc          -- ^ @(Q | R)@
    | Repl Proc              -- ^ @! Q@
    | Let Name Term Proc     -- ^ @let X = T . Q@
    | Match Name Name Proc   -- ^ @[ X = Y ] . Q@
    | Check Proc             -- ^ @chk . Q@
      deriving Show

-- | Address in the process.
newtype Position = Pos String deriving (Show, Eq)

-- | 'Label's for the generated grammar.
newtype Label = Label String  deriving (Show, Eq, Ord)

-- | Productions of the generated grammar.
data Production = P Label Term                -- ^ production
		| Inters Label Label Label    -- ^ intersection rule
                | Remark String               -- ^ comment for the output

-- * Utility

-- | Generate a fresh 'Label' at a given place.
newLabel :: Position -> String -> Label
newLabel (Pos t) s = Label $ t ++ s

-- | 'Label' for a variable at a given place.
newLabelForName :: Position -> Name -> String -> Label
newLabelForName (Pos t) (Name s) extra = Label $ t ++ s ++ extra

-- | Constant 'Term' for a @new@-generated variable.
--   We assume every @new@ to be distinct.
--   (We could also use 'Position' for this, but it would be harder to use the
--   generated name in rewriting rules.)
constOfName :: Name -> String -> Term
constOfName (Name s) extra = App ('k':s ++ extra) []

-- | The top address for the whole proces.
top :: Position
top = Pos "_"

-- | Alternative 'top' (used only in /post-apocalyptic/ analysis).
top' :: Position
top' = Pos "_zapoc_"

-- | Generate a new 'Position' for the continuation.
next :: Position -> Position
next (Pos t) = Pos $ 'n':t

-- | Generate a new 'Position' for the left branch of a 'Par'.
left :: Position -> Position
left (Pos t) = Pos $ 'l':t

-- | Generate a new 'Position' for the right branch of a 'Par'.
right :: Position -> Position
right (Pos t) = Pos $ 'r':t

-- * Pretty printers.

-- | Used for 'Remark's.
pprBinding :: Name -> Label -> String
pprBinding (Name n) (Label l) = n ++ " = @" ++ l ++ " ."

-- | Pretty print a 'Term'.
pprTerm :: Term -> String
pprTerm (App h []) = h
pprTerm (App h ts) = 
    h ++ "(" ++ concat (Data.List.intersperse ", " $ map pprTerm ts) ++ ")"
pprTerm (Lab (Label l)) = '@' : l
pprTerm (Var (Name n)) = n

-- | Pretty print 'Production's.
pprProductions :: [Production] -> String -> String
pprProductions prods comment =
    let isInters :: Production -> Bool
	isInters (Inters _ _ _) = True
	isInters _              = False
        inters, others :: [Production]
	(inters, others) = Data.List.partition isInters prods
        pprInters :: Production -> String
	pprInters (Inters (Label l) (Label l1) (Label l2)) = 
	    "|| @" ++ l ++ " : @" ++ l1 ++ " /\\ @" ++ l2 ++ " .\n"
	pprInters _ = error "pprInters"
        pprOther :: Production -> String
	pprOther (P (Label l) term) = 
	    "@" ++ l ++ " : " ++ pprTerm term ++ " .\n"
	pprOther (Remark note) =
	    "# " ++ note ++ "\n"
	pprOther _ = error "pprOther"
	in "# Intersection rules.\n" ++
           concatMap pprInters inters ++ "\n%%\n\n" ++ comment ++ "\n" ++
	   concatMap pprOther others ++ "\n"

-- | Pretty print a 'Proc'ess.
pprProc :: Proc -> String
pprProc Nil = "()"
pprProc (In (Name n) cont) = "in " ++ n ++ " . " ++ pprProc cont
pprProc (Out term cont) = "out " ++ pprTerm term ++ " . " ++ pprProc cont
pprProc (New (Name n) cont) = "new " ++ n ++ " . " ++ pprProc cont
pprProc (Par p1 p2) =
    let pprPar (Par proc1 proc2) = pprPar proc1 ++ " | " ++ pprPar proc2
        pprPar other             = pprProc other
        in "( " ++ pprPar p1 ++ " | " ++ pprPar p2 ++ " )"
pprProc (Repl proc) = "! " ++ pprProc proc
pprProc (Let (Name n) term cont) = 
    "let " ++ n ++ " = " ++ pprTerm term ++ " . " ++ pprProc cont
pprProc (Match (Name n1) (Name n2) cont) = 
    "[ " ++ n1 ++ " = " ++ n2 ++ " ] . " ++ pprProc cont
pprProc (Check cont) = "chk . " ++ pprProc cont
